# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass, field

from numba_cuda_mlir.numba_cuda import typing
from numba_cuda_mlir.numba_cuda.core import funcdesc, sigutils
from numba_cuda_mlir.numba_cuda.core.callconv import CUDACABICallConv, CUDACallConv
from numba_cuda_mlir.numba_cuda.codegen import ExternalCodeLibrary


class ExternFunction:
    """A descriptor that can be used to call an external device function."""

    def __init__(self, name, sig):
        self.name = name
        self.sig = sig


@dataclass
class DeviceDeclaration:
    name: str
    restype: object
    argtypes: tuple
    link: tuple
    use_cooperative: bool
    abi: str
    extfn: ExternFunction = field(init=False)
    _typing_context_ids: set[int] = field(default_factory=set, init=False)
    _target_context_ids: set[int] = field(default_factory=set, init=False)

    def __post_init__(self):
        self.sig = typing.signature(self.restype, *self.argtypes)
        self.extfn = ExternFunction(self.name, self.sig)

    def apply(self, typingctx, targetctx):
        self.apply_typing(typingctx)
        self.apply_target(targetctx)

    def apply_typing(self, typingctx):
        context_id = id(typingctx)
        if context_id in self._typing_context_ids:
            return
        device_function_template = typing.make_concrete_template(self.name, self.extfn, [self.sig])
        typingctx.insert_user_function(self.extfn, device_function_template)
        self._typing_context_ids.add(context_id)

    def apply_target(self, targetctx):
        context_id = id(targetctx)
        if context_id in self._target_context_ids:
            return

        lib = ExternalCodeLibrary(f"{self.name}_externals", targetctx.codegen())
        for file in self.link:
            lib.add_linking_file(file)
        lib.use_cooperative = self.use_cooperative

        if self.abi == "numba":
            call_conv = CUDACallConv(targetctx)
        elif self.abi == "c":
            call_conv = CUDACABICallConv(targetctx)
        else:
            raise NotImplementedError(f"Unsupported ABI: {self.abi}")

        fndesc = funcdesc.ExternalFunctionDescriptor(
            self.name, self.restype, self.argtypes, call_conv
        )
        targetctx.insert_user_function(self.extfn, fndesc, libs=(lib,))
        self._target_context_ids.add(context_id)


_device_declarations = []


def normalize_device_declaration(name, sig, link=None, use_cooperative=False, abi="numba"):
    if abi not in ("numba", "c"):
        raise NotImplementedError(f"Unsupported ABI: {abi}")

    if link is None:
        link = tuple()
    elif not isinstance(link, (list, tuple, set)):
        link = (link,)
    else:
        link = tuple(link)

    argtypes, restype = sigutils.normalize_signature(sig)
    if restype is None:
        msg = "Return type must be provided for device declarations"
        raise TypeError(msg)

    return name, restype, tuple(argtypes), link, use_cooperative, abi


def register_device_declaration(name, sig, link=None, use_cooperative=False, abi="numba"):
    declaration = DeviceDeclaration(
        *normalize_device_declaration(name, sig, link, use_cooperative, abi)
    )
    _device_declarations.append(declaration)
    return declaration.extfn


def register_device_declaration_from_parts(name, restype, argtypes, link, use_cooperative, abi):
    declaration = DeviceDeclaration(
        name, restype, tuple(argtypes), tuple(link), use_cooperative, abi
    )
    _device_declarations.append(declaration)
    return declaration


_last_applied_index = {}


def apply_device_declarations(typingctx, targetctx):
    key = (id(typingctx), id(targetctx))
    start = _last_applied_index.get(key, 0)
    end = len(_device_declarations)
    if start >= end:
        return
    for declaration in _device_declarations[start:end]:
        declaration.apply(typingctx, targetctx)
    _last_applied_index[key] = end



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from warnings import warn
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.core.errors import (
    DeprecationError,
    NumbaInvalidConfigWarning,
)
from numba_cuda_mlir.numba_cuda.compiler import declare_device_function
from numba_cuda_mlir.numba_cuda.core import sigutils, config
from numba_cuda_mlir.numba_cuda.dispatcher import CUDADispatcher
from numba_cuda_mlir.numba_cuda.cudadrv.driver import _have_nvjitlink


_msg_deprecated_signature_arg = (
    "Deprecated keyword argument `{0}`. "
    "Signatures should be passed as the first "
    "positional argument."
)


def jit(
    func_or_sig=None,
    device=False,
    inline="never",
    forceinline=False,
    link=(),
    debug=None,
    opt=None,
    lineinfo=False,
    cache=False,
    launch_bounds=None,
    lto=None,
    shared_memory_carveout=None,
    **kws,
):
    """
    JIT compile a Python function for CUDA GPUs.

    :param func_or_sig: A function to JIT compile, or *signatures* of a
       function to compile. If a function is supplied, then a
       :class:`Dispatcher <numba.cuda.dispatcher.CUDADispatcher>` is returned.
       Otherwise, ``func_or_sig`` may be a signature or a list of signatures,
       and a function is returned. The returned function accepts another
       function, which it will compile and then return a :class:`Dispatcher
       <numba.cuda.dispatcher.CUDADispatcher>`. See :ref:`jit-decorator` for
       more information about passing signatures.

       .. note:: A kernel cannot have any return value.
    :param device: Indicates whether this is a device function.
    :type device: bool
    :param inline: Enables inlining at the Numba IR level when set to
       ``"always"``. See `Notes on Inlining
       <https://numba.readthedocs.io/en/stable/developer/inlining.html>`_.
    :type inline: str
    :param forceinline: Enables inlining at the NVVM IR level when set to
       ``True``. This is accomplished by adding the ``alwaysinline`` function
       attribute to the function definition.
    :type forceinline: bool
    :param link: A list of files containing PTX or CUDA C/C++ source to link
       with the function
    :type link: list
    :param debug: If True, check for exceptions thrown when executing the
       kernel. Since this degrades performance, this should only be used for
       debugging purposes. If set to True, then ``opt`` should be set to False.
       Defaults to False.  (The default value can be overridden by setting
       environment variable ``NUMBA_CUDA_DEBUGINFO=1``.)
    :param fastmath: When True, enables fastmath optimizations as outlined in
       the :ref:`CUDA Fast Math documentation <cuda-fast-math>`.
    :param max_registers: Request that the kernel is limited to using at most
       this number of registers per thread. The limit may not be respected if
       the ABI requires a greater number of registers than that requested.
       Useful for increasing occupancy.
    :param opt: Whether to compile with optimization enabled. If unspecified,
       the OPT configuration variable is decided by ``NUMBA_OPT```; all
       non-zero values will enable optimization.
    :type opt: bool
    :param lineinfo: If True, generate a line mapping between source code and
       assembly code. This enables inspection of the source code in NVIDIA
       profiling tools and correlation with program counter sampling.
    :type lineinfo: bool
    :param cache: If True, enables the file-based cache for this function.
    :type cache: bool
    :param launch_bounds: Kernel launch bounds, specified as a scalar or a tuple
                          of between one and three items. Tuple items provide:

                          - The maximum number of threads per block,
                          - The minimum number of blocks per SM,
                          - The maximum number of blocks per cluster.

                          If a scalar is provided, it is used as the maximum
                          number of threads per block.
    :type launch_bounds: int | tuple[int]
    :param lto: Whether to enable LTO. If unspecified, LTO is enabled by
                default when nvjitlink is available, except for kernels where
                ``debug=True``.
    :type lto: bool
    :param shared_memory_carveout: Controls the partitioning of shared memory and L1
                                   cache on the GPU. Accepts either a string or an integer:

                                   - String values: ``"MaxL1"`` (maximize L1 cache), ``"MaxShared"``
                                     (maximize shared memory), or ``"default"`` (use driver default).
                                   - Integer values: 0-100 representing the percentage of shared
                                     memory to carve out from the unified memory pool, or -1 for
                                     the default carveout preference.

                                   This parameter is only effective on devices with a unified L1/shared memory
                                   architecture. If unspecified, the CUDA driver uses the default carveout
                                   preference.
    :type shared_memory_carveout: str | int
    """

    if kws.get("boundscheck"):
        raise NotImplementedError("bounds checking is not supported for CUDA")

    if kws.get("argtypes") is not None:
        msg = _msg_deprecated_signature_arg.format("argtypes")
        raise DeprecationError(msg)
    if kws.get("restype") is not None:
        msg = _msg_deprecated_signature_arg.format("restype")
        raise DeprecationError(msg)
    if kws.get("bind") is not None:
        msg = _msg_deprecated_signature_arg.format("bind")
        raise DeprecationError(msg)

    if shared_memory_carveout is not None:
        _validate_shared_memory_carveout(shared_memory_carveout)

    if isinstance(inline, bool):
        DeprecationWarning(
            "Passing bool to inline argument is deprecated, please refer to "
            "Numba's documentation on inlining: "
            "https://numba.readthedocs.io/en/stable/developer/inlining.html. "
            "You may have wanted the forceinline argument instead, to force "
            "inlining at the NVVM IR level."
        )

        inline = "always" if inline else "never"

    debug = config.CUDA_DEBUGINFO_DEFAULT if debug is None else debug
    opt = (config.OPT != 0) if opt is None else opt
    fastmath = kws.get("fastmath", False)
    extensions = kws.get("extensions", [])

    if debug and opt:
        msg = (
            "debug=True with opt=True "
            "is not supported by CUDA. This may result in a crash"
            " - set debug=False or opt=False."
        )
        warn(NumbaInvalidConfigWarning(msg))

    if debug and lineinfo:
        msg = (
            "debug and lineinfo are mutually exclusive. Use debug to get "
            "full debug info (this disables some optimizations), or "
            "lineinfo for line info only with code generation unaffected."
        )
        warn(NumbaInvalidConfigWarning(msg))

    if device and kws.get("link"):
        raise ValueError("link keyword invalid for device function")

    if lto is None:
        # Default to using LTO if nvjitlink is available and we're not debugging
        lto = _have_nvjitlink() and not debug
    else:
        if lto and not _have_nvjitlink():
            raise RuntimeError(
                "LTO requires nvjitlink, which is not availableor not sufficiently recent (>=12.3)"
            )

    if sigutils.is_signature(func_or_sig):
        signatures = [func_or_sig]
        specialized = True
    elif isinstance(func_or_sig, list):
        signatures = func_or_sig
        specialized = False
    else:
        signatures = None

    if signatures is not None:

        def _jit(func):
            targetoptions = kws.copy()
            targetoptions["debug"] = debug
            targetoptions["lineinfo"] = lineinfo
            targetoptions["link"] = link
            targetoptions["opt"] = opt
            targetoptions["fastmath"] = fastmath
            targetoptions["device"] = device
            targetoptions["inline"] = inline
            targetoptions["forceinline"] = forceinline
            targetoptions["extensions"] = extensions
            targetoptions["launch_bounds"] = launch_bounds
            targetoptions["lto"] = lto
            targetoptions["shared_memory_carveout"] = shared_memory_carveout

            disp = CUDADispatcher(func, targetoptions=targetoptions)

            if cache:
                disp.enable_caching()

            for sig in signatures:
                argtypes, restype = sigutils.normalize_signature(sig)

                if restype and not device and restype != types.void:
                    raise TypeError("CUDA kernel must have void return type.")

                if device:
                    from numba_cuda_mlir.numba_cuda.core import typeinfer

                    with typeinfer.register_dispatcher(disp):
                        disp.compile_device(argtypes, restype)
                else:
                    disp.compile(argtypes)

            disp._specialized = specialized
            disp.disable_compile()

            return disp

        return _jit
    else:
        if func_or_sig is None:

            def autojitwrapper(func):
                return jit(
                    func,
                    device=device,
                    inline=inline,
                    forceinline=forceinline,
                    debug=debug,
                    opt=opt,
                    lineinfo=lineinfo,
                    link=link,
                    cache=cache,
                    launch_bounds=launch_bounds,
                    shared_memory_carveout=shared_memory_carveout,
                    **kws,
                )

            return autojitwrapper

        # func_or_sig is a function
        else:
            targetoptions = kws.copy()
            targetoptions["debug"] = debug
            targetoptions["lineinfo"] = lineinfo
            targetoptions["opt"] = opt
            targetoptions["link"] = link
            targetoptions["fastmath"] = fastmath
            targetoptions["device"] = device
            targetoptions["inline"] = inline
            targetoptions["forceinline"] = forceinline
            targetoptions["extensions"] = extensions
            targetoptions["launch_bounds"] = launch_bounds
            targetoptions["lto"] = lto
            targetoptions["shared_memory_carveout"] = shared_memory_carveout
            disp = CUDADispatcher(func_or_sig, targetoptions=targetoptions)

            if cache:
                disp.enable_caching()

            return disp


def declare_device(name, sig, link=None, use_cooperative=False, abi="numba"):
    """
    Declare the signature of a foreign function. Returns a descriptor that can
    be used to call the function from a Python kernel.

    :param name: The name of the foreign function.
    :type name: str
    :param sig: The Numba signature of the function.
    :param link: External code to link when calling the function.
    :param use_cooperative: External code requires cooperative launch.
    :param abi: The ABI to use for the function. "numba" for Numba ABI, "c" for C ABI.
    """
    if abi not in ("numba", "c"):
        raise NotImplementedError(f"Unsupported ABI: {abi}")

    if link is None:
        link = tuple()
    else:
        if not isinstance(link, (list, tuple, set)):
            link = (link,)

    argtypes, restype = sigutils.normalize_signature(sig)
    if restype is None:
        msg = "Return type must be provided for device declarations"
        raise TypeError(msg)

    template = declare_device_function(name, restype, argtypes, link, use_cooperative, abi)

    return template.key


def _validate_shared_memory_carveout(carveout):
    if isinstance(carveout, str):
        valid_strings = ["default", "maxl1", "maxshared"]
        if carveout.lower() not in valid_strings:
            raise ValueError(
                f"Invalid carveout value: {carveout}. Must be -1 to 100 or one of {valid_strings}"
            )
    elif isinstance(carveout, int):
        if not (-1 <= carveout <= 100):
            raise ValueError("Carveout must be between -1 and 100")
    else:
        raise TypeError(f"shared_memory_carveout must be str or int, got {type(carveout).__name__}")



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import os
import platform
import shutil
import pytest
from datetime import datetime
from numba_cuda_mlir.numba_cuda.utils import PYVERSION
from numba_cuda_mlir.numba_cuda.cuda_paths import get_conda_ctk_libdir
from numba_cuda_mlir.numba_cuda.cudadrv import driver, devices, libs
from numba_cuda_mlir.numba_cuda.cudadrv.error import CudaSupportError
from numba_cuda_mlir.numba_cuda.dispatcher import CUDADispatcher
from numba_cuda_mlir.numba_cuda import config
from numba_cuda_mlir.numba_cuda.tests.support import TestCase
from pathlib import Path

from typing import Iterable, Union
from io import StringIO
import unittest
import numpy as np
from numba_cuda_mlir.numba_cuda import HAS_NUMBA

if PYVERSION >= (3, 10):
    from filecheck.matcher import Matcher
    from filecheck.options import Options
    from filecheck.parser import Parser, pattern_for_opts
    from filecheck.finput import FInput

numba_cuda_dir = Path(__file__).parent
test_data_dir = numba_cuda_dir / "tests" / "data"


@pytest.mark.usefixtures("initialize_from_pytest_config")
class CUDATestCase(TestCase):
    """
    For tests that use a CUDA device.

    Methods assertFileCheckAsm and assertFileCheckLLVM will inspect a
    CUDADispatcher and assert that the compilation artifacts match the
    FileCheck checks given in the kernel's docstring.

    Method assertFileCheckMatches can be used to assert that a given string
    matches FileCheck checks, and is not specific to CUDADispatcher.
    """

    FLOAT16_RTOL = np.finfo(np.float16).eps

    def setUp(self):
        self._low_occupancy_warnings = config.CUDA_LOW_OCCUPANCY_WARNINGS
        self._warn_on_implicit_copy = config.CUDA_WARN_ON_IMPLICIT_COPY

        # Disable warnings about low gpu utilization in the test suite
        config.CUDA_LOW_OCCUPANCY_WARNINGS = 0
        # Disable warnings about host arrays in the test suite
        config.CUDA_WARN_ON_IMPLICIT_COPY = 0

    def tearDown(self):
        config.CUDA_LOW_OCCUPANCY_WARNINGS = self._low_occupancy_warnings
        config.CUDA_WARN_ON_IMPLICIT_COPY = self._warn_on_implicit_copy

    Signature = Union[tuple[type, ...], None]

    def _getIRContents(
        self,
        ir_result: Union[dict[Signature, str], str],
        signature: Union[Signature, None] = None,
    ) -> Iterable[str]:
        if isinstance(ir_result, str):
            assert signature is None, (
                "Cannot use signature because the kernel was only compiled for one signature"
            )
            return [ir_result]

        if signature is None:
            return list(ir_result.values())

        return [ir_result[signature]]

    def assertFileCheckAsm(
        self,
        ir_producer: CUDADispatcher,
        signature: Union[tuple[type, ...], None] = None,
        check_prefixes: tuple[str] = ("ASM",),
        **extra_filecheck_options,
    ) -> None:
        """
        Assert that the assembly output of the given CUDADispatcher matches
        the FileCheck checks given in the kernel's docstring.
        """
        ir_contents = self._getIRContents(ir_producer.inspect_asm(), signature)
        assert ir_contents, "No assembly output found for the given signature."
        assert ir_producer.__doc__ is not None, (
            "Kernel docstring is required. To pass checks explicitly, use assertFileCheckMatches."
        )
        check_patterns = ir_producer.__doc__
        for ir_content in ir_contents:
            self.assertFileCheckMatches(
                ir_content,
                check_patterns=check_patterns,
                check_prefixes=check_prefixes,
                **extra_filecheck_options,
            )

    def assertFileCheckLLVM(
        self,
        ir_producer: CUDADispatcher,
        signature: Union[tuple[type, ...], None] = None,
        check_prefixes: tuple[str] = ("LLVM",),
        **extra_filecheck_options,
    ) -> None:
        """
        Assert that the LLVM IR output of the given CUDADispatcher matches
        the FileCheck checks given in the kernel's docstring.
        """
        ir_contents = self._getIRContents(ir_producer.inspect_llvm(), signature)
        assert ir_contents, "No LLVM IR output found for the given signature."
        assert ir_producer.__doc__ is not None, (
            "Kernel docstring is required. To pass checks explicitly, use assertFileCheckMatches."
        )
        check_patterns = ir_producer.__doc__
        for ir_content in ir_contents:
            assert ir_content, "LLVM IR content is empty for the given signature."
            self.assertFileCheckMatches(
                ir_content,
                check_patterns=check_patterns,
                check_prefixes=check_prefixes,
                **extra_filecheck_options,
            )

    def assertFileCheckMatches(
        self,
        ir_content: str,
        check_patterns: str,
        check_prefixes: tuple[str] = ("CHECK",),
        **extra_filecheck_options,
    ) -> None:
        """
        Assert that the given string matches the passed FileCheck checks.

        Args:
            ir_content: The string to check against.
            check_patterns: The FileCheck checks to use.
            check_prefixes: The prefixes to use for the FileCheck checks.
            extra_filecheck_options: Extra options to pass to FileCheck.
        """
        if PYVERSION < (3, 10):
            self.skipTest("FileCheck requires Python 3.10 or later")
        opts = Options(
            match_filename="-",
            check_prefixes=list(check_prefixes),
            **extra_filecheck_options,
        )
        input_file = FInput(fname="-", content=ir_content)
        parser = Parser(opts, StringIO(check_patterns), *pattern_for_opts(opts))
        matcher = Matcher(opts, input_file, parser)
        matcher.stderr = StringIO()
        result = matcher.run()
        if result != 0:
            if self._dump_failed_filechecks:
                dump_directory = Path(datetime.now().strftime("numba-ir-%Y_%m_%d_%H_%M_%S"))
                if not dump_directory.exists():
                    dump_directory.mkdir(parents=True, exist_ok=True)
                base_path = self.id().replace(".", "_")
                ir_dump = dump_directory / Path(base_path).with_suffix(".ll")
                checks_dump = dump_directory / Path(base_path).with_suffix(".checks")
                with (
                    open(ir_dump, "w") as ir_file,
                    open(checks_dump, "w") as checks_file,
                ):
                    _ = ir_file.write(ir_content + "\n")
                    _ = checks_file.write(check_patterns)
                    dump_instructions = f"Reproduce with:\n\nfilecheck --check-prefixes={','.join(check_prefixes)} {checks_dump} --input-file {ir_dump}"
            else:
                dump_instructions = "Rerun with --dump-failed-filechecks to generate a reproducer."

            self.fail(f"FileCheck failed:\n{matcher.stderr.getvalue()}\n\n" + dump_instructions)


skip_on_standalone_numba_cuda = unittest.skipUnless(HAS_NUMBA, "requires base numba install")


def skip_unless_conda_cudatoolkit(reason):
    """Skip test if the CUDA toolkit was not installed by Conda"""
    assert isinstance(reason, str)
    return unittest.skipUnless(get_conda_ctk_libdir() is not None, reason)


def skip_if_external_memmgr(reason):
    """Skip test if an EMM Plugin is in use"""
    assert isinstance(reason, str)
    return unittest.skipIf(config.CUDA_MEMORY_MANAGER != "default", reason)


def skip_under_cuda_memcheck(reason):
    assert isinstance(reason, str)
    return unittest.skipIf(os.environ.get("CUDA_MEMCHECK") is not None, reason)


def skip_without_nvdisasm(reason):
    assert isinstance(reason, str)
    nvdisasm_path = shutil.which("nvdisasm")
    return unittest.skipIf(nvdisasm_path is None, reason)


def skip_with_nvdisasm(reason):
    assert isinstance(reason, str)
    nvdisasm_path = shutil.which("nvdisasm")
    return unittest.skipIf(nvdisasm_path is not None, reason)


def skip_on_arm(reason):
    assert isinstance(reason, str)
    cpu = platform.processor()
    is_arm = cpu.startswith("arm") or cpu.startswith("aarch")
    return unittest.skipIf(is_arm, reason)


def skip_on_wsl2(reason):
    """Skip test when running under WSL2.

    Detection is based on the kernel release string, which typically contains
    "microsoft-standard-WSL2" on WSL2 systems.
    """
    assert isinstance(reason, str)
    rel = platform.release().lower()
    is_wsl2 = ("microsoft-standard-wsl2" in rel) or ("wsl2" in rel)
    return unittest.skipIf(is_wsl2, reason)


def skip_if_cuda_includes_missing(fn):
    # Skip when cuda.h is not available - generally this should indicate
    # whether the CUDA includes are available or not
    reason = "CUDA include dir not available on this system"
    try:
        cuda_include_path = libs.get_cuda_include_dir()
    except FileNotFoundError:
        return unittest.skip(reason)(fn)
    cuda_h = os.path.join(cuda_include_path, "cuda.h")
    cuda_h_file = os.path.exists(cuda_h) and os.path.isfile(cuda_h)
    return unittest.skipUnless(cuda_h_file, reason)(fn)


def skip_if_curand_kernel_missing(fn):
    reason = "curand_kernel.h not available on this system"
    try:
        cuda_include_path = libs.get_cuda_include_dir()
    except FileNotFoundError:
        return unittest.skip(reason)(fn)
    curand_kernel_h = os.path.join(cuda_include_path, "curand_kernel.h")
    curand_kernel_h_file = os.path.exists(curand_kernel_h) and os.path.isfile(curand_kernel_h)
    return unittest.skipUnless(curand_kernel_h_file, reason)(fn)


def _get_device_compute_capability():
    try:
        return devices.gpus[0].compute_capability
    except (CudaSupportError, RuntimeError, ValueError, IndexError):
        return None


def cc_X_or_above(major, minor):
    cc = _get_device_compute_capability()
    return cc is not None and cc >= (major, minor)


def skip_unless_cc_50(fn):
    return unittest.skipUnless(cc_X_or_above(5, 0), "requires cc >= 5.0")(fn)


def skip_unless_cc_53(fn):
    return unittest.skipUnless(cc_X_or_above(5, 3), "requires cc >= 5.3")(fn)


def skip_unless_cc_60(fn):
    return unittest.skipUnless(cc_X_or_above(6, 0), "requires cc >= 6.0")(fn)


def skip_unless_cc_75(fn):
    return unittest.skipUnless(cc_X_or_above(7, 5), "requires cc >= 7.5")(fn)


def skip_unless_cc_90(fn):
    return unittest.skipUnless(cc_X_or_above(9, 0), "requires cc >= 9.0")(fn)


def cudadevrt_missing():
    try:
        path = libs.get_cudalib("cudadevrt", static=True)
        libs.check_static_lib(path)
    except FileNotFoundError:
        return True
    return False


def skip_if_cudadevrt_missing(fn):
    return unittest.skipIf(cudadevrt_missing(), "cudadevrt missing")(fn)


def skip_if_nvjitlink_missing(reason):
    assert isinstance(reason, str)
    return unittest.skipIf(not driver._have_nvjitlink(), reason)


def _is_nvjitlink_13_1_and_sm_120():
    """Check if nvjitlink version is 13.1 and compute capability is 120 (sm_120).

    sm_120 refers to compute capability 12.0, represented as the tuple (12, 0).
    """
    try:
        from cuda.bindings import nvjitlink

        nvjitlink_ver = nvjitlink.version()
        # Check if nvjitlink version is 13.1.x
        if nvjitlink_ver[0] != 13 or nvjitlink_ver[1] != 1:
            return False

        # Check if compute capability is 12.0 (sm_120)
        cc = _get_device_compute_capability()
        if cc != (12, 0):
            return False

        return True
    except (ImportError, AttributeError, RuntimeError):
        # ImportError: nvjitlink not available
        # AttributeError: version() method missing
        # RuntimeError: device context issues
        return False


def skip_on_nvjitlink_13_1_sm_120(reason):
    """Skip test when nvjitlink version is 13.1 and compute capability is sm_120.

    This is used to skip tests that fail at link time with nvjitlink 13.1 on sm_120
    GPUs (e.g., tests calling sum, mean, etc. in numba kernels).
    """
    assert isinstance(reason, str)
    return unittest.skipIf(_is_nvjitlink_13_1_and_sm_120(), reason)


class ForeignArray:
    """
    Class for emulating an array coming from another library through the CUDA
    Array interface. This just hides a DeviceNDArray so that it doesn't look
    like a DeviceNDArray.
    """

    def __init__(self, arr):
        self._arr = arr
        self.__cuda_array_interface__ = arr.__cuda_array_interface__



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Complex math intrinsics with IEEE 754 special case handling.
"""

from typing import TypeVar
from numba_cuda_mlir._mlir.dialects import math as math_dialect
from numba_cuda_mlir.mlir.dialect_exts import func, scf, arith
from numba_cuda_mlir.mlir.context import mlir_mod_ctx
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir.mlir.dialect_exts.scf import (
    if_ctx_manager as if_,
    else_ctx_manager as else_,
)
from numba_cuda_mlir._mlir.ir import UnitAttr

FTy = TypeVar("FTy")


def _nan(fty):
    return arith.constant(float("nan"), fty)


def _inf(fty):
    return arith.constant(float("inf"), fty)


def _neg_inf(fty):
    return arith.constant(float("-inf"), fty)


def _zero(fty):
    return arith.constant(0.0, fty)


def _neg_zero(fty):
    return arith.constant(-0.0, fty)


@func.func(sym_visibility="private", generics=[FTy])
def cmath_exp(z_real: FTy, z_imag: FTy):
    c0 = _zero(FTy)
    nan = _nan(FTy)
    neg_inf = _neg_inf(FTy)

    x_is_finite = math_dialect.isfinite(z_real)
    y_is_finite = math_dialect.isfinite(z_imag)
    x_isnan = math_dialect.isnan(z_real)
    x_is_neg_inf = arith.cmpf(arith.CmpFPredicate.OEQ, z_real, neg_inf)
    y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, z_imag, c0)

    with if_(x_is_finite, results=[FTy, FTy]) as result:
        with if_(y_is_finite, results=[FTy, FTy]) as result_inner:
            exp_x = math_dialect.exp(z_real)
            cos_y = math_dialect.cos(z_imag)
            sin_y = math_dialect.sin(z_imag)
            real_part = arith.mulf(exp_x, cos_y)
            imag_part = arith.mulf(exp_x, sin_y)
            scf.yield_(real_part, imag_part)
        with else_(result_inner):
            scf.yield_(nan, nan)
        scf.yield_(result_inner.results)
    with else_(result):
        with if_(x_isnan, results=[FTy, FTy]) as result_outer:
            with if_(y_is_zero, results=[FTy, FTy]) as result_nan:
                scf.yield_(z_real, z_imag)
            with else_(result_nan):
                scf.yield_(nan, nan)
            scf.yield_(result_nan.results)
        with else_(result_outer):
            with if_(x_is_neg_inf, results=[FTy, FTy]) as result_inf:
                with if_(y_is_finite, results=[FTy, FTy]) as result_neg_inf_yfin:
                    cos_y_neg = math_dialect.cos(z_imag)
                    sin_y_neg = math_dialect.sin(z_imag)
                    real_part_neg = math_dialect.copysign(c0, cos_y_neg)
                    imag_part_neg = math_dialect.copysign(c0, sin_y_neg)
                    scf.yield_(real_part_neg, imag_part_neg)
                with else_(result_neg_inf_yfin):
                    scf.yield_(c0, c0)
                scf.yield_(result_neg_inf_yfin.results)
            with else_(result_inf):
                with if_(y_is_finite, results=[FTy, FTy]) as result_pos_inf:
                    cos_y = math_dialect.cos(z_imag)
                    sin_y = math_dialect.sin(z_imag)
                    cos_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, cos_y, c0)
                    sin_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, sin_y, c0)
                    real_mul = arith.mulf(z_real, cos_y)
                    imag_mul = arith.mulf(z_real, sin_y)
                    real_part = arith.select(cos_is_zero, cos_y, real_mul)
                    imag_part = arith.select(sin_is_zero, sin_y, imag_mul)
                    scf.yield_(real_part, imag_part)
                with else_(result_pos_inf):
                    scf.yield_(z_real, nan)
                scf.yield_(result_pos_inf.results)
            scf.yield_(result_inf.results)
        scf.yield_(result_outer.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_sinh(z_real: FTy, z_imag: FTy):
    c0 = _zero(FTy)
    nan = _nan(FTy)

    x_isinf = math_dialect.isinf(z_real)
    y_isnan = math_dialect.isnan(z_imag)
    y_is_finite = math_dialect.isfinite(z_imag)
    x_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, z_real, c0)

    with if_(x_isinf, results=[FTy, FTy]) as result:
        with if_(y_isnan, results=[FTy, FTy]) as result_inf:
            scf.yield_(z_real, z_imag)
        with else_(result_inf):
            cos_y = math_dialect.cos(z_imag)
            sin_y = math_dialect.sin(z_imag)
            cos_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, cos_y, c0)
            sin_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, sin_y, c0)
            abs_x = math_dialect.absf(z_real)
            real_mul = arith.mulf(z_real, cos_y)
            imag_mul = arith.mulf(abs_x, sin_y)
            real_part = arith.select(cos_is_zero, cos_y, real_mul)
            imag_part = arith.select(sin_is_zero, sin_y, imag_mul)
            scf.yield_(real_part, imag_part)
        scf.yield_(result_inf.results)
    with else_(result):
        with if_(y_is_finite, results=[FTy, FTy]) as result_finite:
            sinh_x = math_dialect.sinh(z_real)
            cosh_x = math_dialect.cosh(z_real)
            cos_y = math_dialect.cos(z_imag)
            sin_y = math_dialect.sin(z_imag)
            real_part = arith.mulf(sinh_x, cos_y)
            imag_part = arith.mulf(cosh_x, sin_y)
            scf.yield_(real_part, imag_part)
        with else_(result_finite):
            with if_(x_is_zero, results=[FTy, FTy]) as result_yinf:
                scf.yield_(z_real, nan)
            with else_(result_yinf):
                scf.yield_(nan, nan)
            scf.yield_(result_yinf.results)
        scf.yield_(result_finite.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_cosh(z_real: FTy, z_imag: FTy):
    c0 = _zero(FTy)
    nan = _nan(FTy)

    x_isinf = math_dialect.isinf(z_real)
    x_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, z_real, c0)
    y_isnan = math_dialect.isnan(z_imag)
    y_is_finite = math_dialect.isfinite(z_imag)
    y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, z_imag, c0)

    with if_(x_isinf, results=[FTy, FTy]) as result:
        abs_x = math_dialect.absf(z_real)
        with if_(y_isnan, results=[FTy, FTy]) as result_inf:
            scf.yield_(abs_x, z_imag)
        with else_(result_inf):
            with if_(y_is_zero, results=[FTy, FTy]) as result_not_nan:
                one = arith.constant(1.0, FTy)
                x_sign = math_dialect.copysign(one, z_real)
                y_sign = math_dialect.copysign(one, z_imag)
                sign_product = arith.mulf(x_sign, y_sign)
                imag_with_sign = math_dialect.copysign(c0, sign_product)
                scf.yield_(abs_x, imag_with_sign)
            with else_(result_not_nan):
                cos_y = math_dialect.cos(z_imag)
                sin_y = math_dialect.sin(z_imag)
                real_sign = math_dialect.copysign(z_real, cos_y)
                imag_sign = math_dialect.copysign(z_real, sin_y)
                x_is_neg = arith.cmpf(arith.CmpFPredicate.OLT, z_real, c0)
                neg_imag_sign = arith.negf(imag_sign)
                imag_part = arith.select(x_is_neg, neg_imag_sign, imag_sign)
                scf.yield_(real_sign, imag_part)
            scf.yield_(result_not_nan.results)
        scf.yield_(result_inf.results)
    with else_(result):
        with if_(y_is_finite, results=[FTy, FTy]) as result_finite:
            cosh_x = math_dialect.cosh(z_real)
            sinh_x = math_dialect.sinh(z_real)
            cos_y = math_dialect.cos(z_imag)
            sin_y = math_dialect.sin(z_imag)
            real_part = arith.mulf(cosh_x, cos_y)
            imag_part = arith.mulf(sinh_x, sin_y)
            scf.yield_(real_part, imag_part)
        with else_(result_finite):
            with if_(x_is_zero, results=[FTy, FTy]) as result_yinf:
                scf.yield_(nan, c0)
            with else_(result_yinf):
                scf.yield_(nan, nan)
            scf.yield_(result_yinf.results)
        scf.yield_(result_finite.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_sin(z_real: FTy, z_imag: FTy):
    c0 = _zero(FTy)
    nan = _nan(FTy)
    inf = _inf(FTy)

    x_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, z_real, c0)
    y_isinf = math_dialect.isinf(z_imag)
    y_isnan = math_dialect.isnan(z_imag)
    x_is_finite = math_dialect.isfinite(z_real)
    y_is_finite = math_dialect.isfinite(z_imag)

    pure_imag_inf = arith.andi(x_is_zero, y_isinf)
    with if_(pure_imag_inf, results=[FTy, FTy]) as result:
        scf.yield_(z_real, z_imag)
    with else_(result):
        pure_imag_nan = arith.andi(x_is_zero, y_isnan)
        with if_(pure_imag_nan, results=[FTy, FTy]) as result_pure:
            scf.yield_(z_real, z_imag)
        with else_(result_pure):
            with if_(x_is_finite, results=[FTy, FTy]) as result_normal:
                with if_(y_is_finite, results=[FTy, FTy]) as result_xfin:
                    sin_x = math_dialect.sin(z_real)
                    cos_x = math_dialect.cos(z_real)
                    sinh_y = math_dialect.sinh(z_imag)
                    cosh_y = math_dialect.cosh(z_imag)
                    real_part = arith.mulf(sin_x, cosh_y)
                    imag_part = arith.mulf(cos_x, sinh_y)
                    scf.yield_(real_part, imag_part)
                with else_(result_xfin):
                    with if_(y_isinf, results=[FTy, FTy]) as result_ynotfin:
                        sin_x = math_dialect.sin(z_real)
                        cos_x = math_dialect.cos(z_real)
                        real_part = math_dialect.copysign(inf, sin_x)
                        imag_sign = arith.mulf(cos_x, z_imag)
                        imag_part = math_dialect.copysign(inf, imag_sign)
                        scf.yield_(real_part, imag_part)
                    with else_(result_ynotfin):
                        scf.yield_(nan, nan)
                    scf.yield_(result_ynotfin.results)
                scf.yield_(result_xfin.results)
            with else_(result_normal):
                with if_(y_is_finite, results=[FTy, FTy]) as result_xinf:
                    y_is_zero_inner = arith.cmpf(arith.CmpFPredicate.OEQ, z_imag, c0)
                    with if_(y_is_zero_inner, results=[FTy, FTy]) as result_yzero:
                        scf.yield_(nan, z_imag)
                    with else_(result_yzero):
                        scf.yield_(nan, nan)
                    scf.yield_(result_yzero.results)
                with else_(result_xinf):
                    scf.yield_(nan, nan)
                scf.yield_(result_xinf.results)
            scf.yield_(result_normal.results)
        scf.yield_(result_pure.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_cos(z_real: FTy, z_imag: FTy):
    c0 = _zero(FTy)
    nan = _nan(FTy)

    x_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, z_real, c0)
    x_is_finite = math_dialect.isfinite(z_real)
    y_is_finite = math_dialect.isfinite(z_imag)

    with if_(x_is_finite, results=[FTy, FTy]) as result:
        with if_(y_is_finite, results=[FTy, FTy]) as result_xfin:
            cos_x = math_dialect.cos(z_real)
            sin_x = math_dialect.sin(z_real)
            cosh_y = math_dialect.cosh(z_imag)
            sinh_y = math_dialect.sinh(z_imag)
            real_part = arith.mulf(cos_x, cosh_y)
            sin_sinh = arith.mulf(sin_x, sinh_y)
            imag_part = arith.negf(sin_sinh)
            scf.yield_(real_part, imag_part)
        with else_(result_xfin):
            y_isnan = math_dialect.isnan(z_imag)
            with if_(y_isnan, results=[FTy, FTy]) as result_ynan:
                scf.yield_(nan, nan)
            with else_(result_ynan):
                with if_(x_is_zero, results=[FTy, FTy]) as result_yinf:
                    abs_y = math_dialect.absf(z_imag)
                    one = arith.constant(1.0, FTy)
                    x_sign = math_dialect.copysign(one, z_real)
                    y_sign = math_dialect.copysign(one, z_imag)
                    neg_x_sign = arith.negf(x_sign)
                    imag_sign = arith.mulf(neg_x_sign, y_sign)
                    imag_zero = math_dialect.copysign(c0, imag_sign)
                    scf.yield_(abs_y, imag_zero)
                with else_(result_yinf):
                    inf_val = _inf(FTy)
                    cos_x = math_dialect.cos(z_real)
                    sin_x = math_dialect.sin(z_real)
                    neg_sin_x = arith.negf(sin_x)
                    imag_sign = arith.mulf(neg_sin_x, z_imag)
                    real_part_yinf = math_dialect.copysign(inf_val, cos_x)
                    imag_part_yinf = math_dialect.copysign(inf_val, imag_sign)
                    scf.yield_(real_part_yinf, imag_part_yinf)
                scf.yield_(result_yinf.results)
            scf.yield_(result_ynan.results)
        scf.yield_(result_xfin.results)
    with else_(result):
        with if_(y_is_finite, results=[FTy, FTy]) as result_xinf:
            y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, z_imag, c0)
            with if_(y_is_zero, results=[FTy, FTy]) as result_yzero:
                scf.yield_(nan, z_imag)
            with else_(result_yzero):
                scf.yield_(nan, nan)
            scf.yield_(result_yzero.results)
        with else_(result_xinf):
            scf.yield_(nan, nan)
        scf.yield_(result_xinf.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_rect(r: FTy, phi: FTy):
    c0 = _zero(FTy)
    nan = _nan(FTy)

    r_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, r, c0)
    r_isinf = math_dialect.isinf(r)
    phi_is_finite = math_dialect.isfinite(phi)

    with if_(phi_is_finite, results=[FTy, FTy]) as result:
        cos_phi = math_dialect.cos(phi)
        sin_phi = math_dialect.sin(phi)
        with if_(r_isinf, results=[FTy, FTy]) as result_finite:
            cos_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, cos_phi, c0)
            sin_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, sin_phi, c0)
            real_mul = arith.mulf(r, cos_phi)
            imag_mul = arith.mulf(r, sin_phi)
            one = arith.constant(1.0, FTy)
            r_sign = math_dialect.copysign(one, r)
            cos_sign = math_dialect.copysign(one, cos_phi)
            sin_sign = math_dialect.copysign(one, sin_phi)
            real_sign = arith.mulf(r_sign, cos_sign)
            imag_sign = arith.mulf(r_sign, sin_sign)
            real_zero = math_dialect.copysign(c0, real_sign)
            imag_zero = math_dialect.copysign(c0, imag_sign)
            real_part = arith.select(cos_is_zero, real_zero, real_mul)
            imag_part = arith.select(sin_is_zero, imag_zero, imag_mul)
            scf.yield_(real_part, imag_part)
        with else_(result_finite):
            real_part = arith.mulf(r, cos_phi)
            imag_part = arith.mulf(r, sin_phi)
            scf.yield_(real_part, imag_part)
        scf.yield_(result_finite.results)
    with else_(result):
        with if_(r_is_zero, results=[FTy, FTy]) as result_phi_inf:
            abs_r = math_dialect.absf(r)
            scf.yield_(abs_r, c0)
        with else_(result_phi_inf):
            with if_(r_isinf, results=[FTy, FTy]) as result_r_nonzero:
                scf.yield_(r, nan)
            with else_(result_r_nonzero):
                scf.yield_(nan, nan)
            scf.yield_(result_r_nonzero.results)
        scf.yield_(result_phi_inf.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_sqrt(z_real: FTy, z_imag: FTy):
    c0 = _zero(FTy)
    nan = _nan(FTy)
    inf = _inf(FTy)
    neg_inf = _neg_inf(FTy)
    half = arith.constant(0.5, FTy)
    two = arith.constant(2.0, FTy)

    x = z_real
    y = z_imag

    x_isinf = math_dialect.isinf(x)
    y_isinf = math_dialect.isinf(y)
    x_isnan = math_dialect.isnan(x)
    y_isnan = math_dialect.isnan(y)
    x_is_pos_inf = arith.cmpf(arith.CmpFPredicate.OEQ, x, inf)
    x_is_neg_inf = arith.cmpf(arith.CmpFPredicate.OEQ, x, neg_inf)

    with if_(y_isinf, results=[FTy, FTy]) as result:
        scf.yield_(inf, y)
    with else_(result):
        with if_(x_is_neg_inf, results=[FTy, FTy]) as result_xinf:
            with if_(y_isnan, results=[FTy, FTy]) as result_ynan:
                scf.yield_(y, y)
            with else_(result_ynan):
                imag_res = math_dialect.copysign(inf, y)
                scf.yield_(c0, imag_res)
            scf.yield_(result_ynan.results)
        with else_(result_xinf):
            with if_(x_is_pos_inf, results=[FTy, FTy]) as result_xposinf:
                with if_(y_isnan, results=[FTy, FTy]) as result_ynan2:
                    scf.yield_(inf, y)
                with else_(result_ynan2):
                    imag_res = math_dialect.copysign(c0, y)
                    scf.yield_(inf, imag_res)
                scf.yield_(result_ynan2.results)
            with else_(result_xposinf):
                with if_(x_isnan, results=[FTy, FTy]) as result_xnan:
                    scf.yield_(nan, nan)
                with else_(result_xnan):
                    with if_(y_isnan, results=[FTy, FTy]) as result_ynan3:
                        scf.yield_(nan, nan)
                    with else_(result_ynan3):
                        abs_x = math_dialect.absf(x)
                        abs_y = math_dialect.absf(y)
                        x2 = arith.mulf(x, x)
                        y2 = arith.mulf(y, y)
                        modulus = math_dialect.sqrt(arith.addf(x2, y2))
                        t = math_dialect.sqrt(arith.mulf(half, arith.addf(abs_x, modulus)))
                        x_is_neg = arith.cmpf(arith.CmpFPredicate.OLT, x, c0)
                        two_t = arith.mulf(two, t)
                        y_div_2t = arith.divf(y, two_t)
                        abs_y_div_2t = arith.divf(abs_y, two_t)
                        t_signed = math_dialect.copysign(t, y)
                        real_res = arith.select(x_is_neg, abs_y_div_2t, t)
                        imag_res = arith.select(x_is_neg, t_signed, y_div_2t)
                        x_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, x, c0)
                        y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, y, c0)
                        both_zero = arith.andi(x_is_zero, y_is_zero)
                        zero_imag = math_dialect.copysign(c0, y)
                        final_real = arith.select(both_zero, c0, real_res)
                        final_imag = arith.select(both_zero, zero_imag, imag_res)
                        scf.yield_(final_real, final_imag)
                    scf.yield_(result_ynan3.results)
                scf.yield_(result_xnan.results)
            scf.yield_(result_xposinf.results)
        scf.yield_(result_xinf.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_acos(z_real: FTy, z_imag: FTy):
    c0 = _zero(FTy)
    c1 = arith.constant(1.0, FTy)
    nan = _nan(FTy)
    inf = _inf(FTy)
    pi = arith.constant(3.141592653589793, FTy)
    pi_over_2 = arith.constant(1.5707963267948966, FTy)

    x = z_real
    y = z_imag

    x_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, x, c0)
    y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, y, c0)
    x_isinf = math_dialect.isinf(x)
    y_isinf = math_dialect.isinf(y)
    x_isnan = math_dialect.isnan(x)
    y_isnan = math_dialect.isnan(y)
    x_is_neg = arith.cmpf(arith.CmpFPredicate.OLT, x, c0)

    both_zero = arith.andi(x_is_zero, y_is_zero)
    with if_(both_zero, results=[FTy, FTy]) as result:
        neg_y = arith.negf(y)
        scf.yield_(pi_over_2, neg_y)
    with else_(result):
        with if_(y_isinf, results=[FTy, FTy]) as result_yinf:
            imag_res = arith.negf(y)
            scf.yield_(pi_over_2, imag_res)
        with else_(result_yinf):
            with if_(x_isinf, results=[FTy, FTy]) as result_xinf:
                with if_(y_isnan, results=[FTy, FTy]) as result_ynan:
                    scf.yield_(nan, nan)
                with else_(result_ynan):
                    neg_y_inf = arith.negf(y)
                    imag_pos_inf = math_dialect.copysign(c0, neg_y_inf)
                    imag_neg_inf = math_dialect.copysign(inf, neg_y_inf)
                    real_res = arith.select(x_is_neg, pi, c0)
                    imag_res = arith.select(x_is_neg, imag_neg_inf, imag_pos_inf)
                    scf.yield_(real_res, imag_res)
                scf.yield_(result_ynan.results)
            with else_(result_xinf):
                with if_(x_isnan, results=[FTy, FTy]) as result_xnan:
                    scf.yield_(nan, nan)
                with else_(result_xnan):
                    with if_(y_isnan, results=[FTy, FTy]) as result_ynan2:
                        scf.yield_(nan, nan)
                    with else_(result_ynan2):
                        x2 = arith.mulf(x, x)
                        y2 = arith.mulf(y, y)
                        one_minus_x2_plus_y2 = arith.subf(c1, arith.subf(x2, y2))
                        two_xy = arith.mulf(arith.constant(2.0, FTy), arith.mulf(x, y))
                        neg_two_xy = arith.negf(two_xy)
                        mod_sq = arith.addf(
                            arith.mulf(one_minus_x2_plus_y2, one_minus_x2_plus_y2),
                            arith.mulf(neg_two_xy, neg_two_xy),
                        )
                        mod = math_dialect.sqrt(mod_sq)
                        phase = math_dialect.atan2(neg_two_xy, one_minus_x2_plus_y2)
                        sqrt_mod = math_dialect.sqrt(mod)
                        half_phase = arith.mulf(arith.constant(0.5, FTy), phase)
                        sqrt_real = arith.mulf(sqrt_mod, math_dialect.cos(half_phase))
                        sqrt_imag = arith.mulf(sqrt_mod, math_dialect.sin(half_phase))
                        sum_real = arith.subf(x, sqrt_imag)
                        sum_imag = arith.addf(y, sqrt_real)
                        log_mod = math_dialect.log(
                            math_dialect.sqrt(
                                arith.addf(
                                    arith.mulf(sum_real, sum_real),
                                    arith.mulf(sum_imag, sum_imag),
                                )
                            )
                        )
                        log_phase = math_dialect.atan2(sum_imag, sum_real)
                        final_real = log_phase
                        final_imag = arith.negf(log_mod)
                        scf.yield_(final_real, final_imag)
                    scf.yield_(result_ynan2.results)
                scf.yield_(result_xnan.results)
            scf.yield_(result_xinf.results)
        scf.yield_(result_yinf.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_acosh(z_real: FTy, z_imag: FTy):
    c0 = _zero(FTy)
    c1 = arith.constant(1.0, FTy)
    nan = _nan(FTy)
    inf = _inf(FTy)
    pi = arith.constant(3.141592653589793, FTy)
    pi_over_2 = arith.constant(1.5707963267948966, FTy)

    x = z_real
    y = z_imag

    x_isinf = math_dialect.isinf(x)
    y_isinf = math_dialect.isinf(y)
    x_isnan = math_dialect.isnan(x)
    y_isnan = math_dialect.isnan(y)
    x_is_neg = arith.cmpf(arith.CmpFPredicate.OLT, x, c0)
    y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, y, c0)
    x_is_one = arith.cmpf(arith.CmpFPredicate.OEQ, x, c1)

    one_with_zero = arith.andi(x_is_one, y_is_zero)
    with if_(one_with_zero, results=[FTy, FTy]) as result:
        scf.yield_(c0, y)
    with else_(result):
        with if_(y_isinf, results=[FTy, FTy]) as result_yinf:
            imag_res = math_dialect.copysign(pi_over_2, y)
            scf.yield_(inf, imag_res)
        with else_(result_yinf):
            with if_(x_isinf, results=[FTy, FTy]) as result_xinf:
                with if_(y_isnan, results=[FTy, FTy]) as result_ynan:
                    scf.yield_(inf, nan)
                with else_(result_ynan):
                    imag_pos = math_dialect.copysign(c0, y)
                    imag_neg = math_dialect.copysign(pi, y)
                    imag_res = arith.select(x_is_neg, imag_neg, imag_pos)
                    scf.yield_(inf, imag_res)
                scf.yield_(result_ynan.results)
            with else_(result_xinf):
                with if_(x_isnan, results=[FTy, FTy]) as result_xnan:
                    scf.yield_(nan, nan)
                with else_(result_xnan):
                    with if_(y_isnan, results=[FTy, FTy]) as result_ynan2:
                        scf.yield_(nan, nan)
                    with else_(result_ynan2):
                        x2 = arith.mulf(x, x)
                        y2 = arith.mulf(y, y)
                        two_xy = arith.mulf(arith.constant(2.0, FTy), arith.mulf(x, y))
                        z2_minus_1_real = arith.subf(arith.subf(x2, y2), c1)
                        z2_minus_1_imag = two_xy
                        mod_sq = arith.addf(
                            arith.mulf(z2_minus_1_real, z2_minus_1_real),
                            arith.mulf(z2_minus_1_imag, z2_minus_1_imag),
                        )
                        mod = math_dialect.sqrt(mod_sq)
                        phase = math_dialect.atan2(z2_minus_1_imag, z2_minus_1_real)
                        sqrt_mod = math_dialect.sqrt(mod)
                        half_phase = arith.mulf(arith.constant(0.5, FTy), phase)
                        sqrt_real = arith.mulf(sqrt_mod, math_dialect.cos(half_phase))
                        sqrt_imag = arith.mulf(sqrt_mod, math_dialect.sin(half_phase))
                        sum_real = arith.addf(x, sqrt_real)
                        sum_imag = arith.addf(y, sqrt_imag)
                        log_mod = math_dialect.log(
                            math_dialect.sqrt(
                                arith.addf(
                                    arith.mulf(sum_real, sum_real),
                                    arith.mulf(sum_imag, sum_imag),
                                )
                            )
                        )
                        log_phase = math_dialect.atan2(sum_imag, sum_real)
                        abs_log_mod = math_dialect.absf(log_mod)
                        abs_log_phase = math_dialect.absf(log_phase)
                        final_imag = math_dialect.copysign(abs_log_phase, y)
                        scf.yield_(abs_log_mod, final_imag)
                    scf.yield_(result_ynan2.results)
                scf.yield_(result_xnan.results)
            scf.yield_(result_xinf.results)
        scf.yield_(result_yinf.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_atan(z_real: FTy, z_imag: FTy):
    c0 = _zero(FTy)
    c1 = arith.constant(1.0, FTy)
    nan = _nan(FTy)
    inf = _inf(FTy)
    pi_over_2 = arith.constant(1.5707963267948966, FTy)

    x = z_real
    y = z_imag

    x_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, x, c0)
    y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, y, c0)
    x_isinf = math_dialect.isinf(x)
    y_isinf = math_dialect.isinf(y)
    x_isnan = math_dialect.isnan(x)
    y_isnan = math_dialect.isnan(y)

    both_zero = arith.andi(x_is_zero, y_is_zero)
    with if_(both_zero, results=[FTy, FTy]) as result:
        scf.yield_(x, y)
    with else_(result):
        with if_(y_isinf, results=[FTy, FTy]) as result_yinf:
            real_res = math_dialect.copysign(pi_over_2, x)
            imag_res = math_dialect.copysign(c0, y)
            scf.yield_(real_res, imag_res)
        with else_(result_yinf):
            with if_(x_isinf, results=[FTy, FTy]) as result_xinf:
                with if_(y_isnan, results=[FTy, FTy]) as result_ynan:
                    real_res = math_dialect.copysign(pi_over_2, x)
                    scf.yield_(real_res, nan)
                with else_(result_ynan):
                    real_res = math_dialect.copysign(pi_over_2, x)
                    imag_res = math_dialect.copysign(c0, y)
                    scf.yield_(real_res, imag_res)
                scf.yield_(result_ynan.results)
            with else_(result_xinf):
                with if_(x_isnan, results=[FTy, FTy]) as result_xnan:
                    scf.yield_(nan, nan)
                with else_(result_xnan):
                    with if_(y_isnan, results=[FTy, FTy]) as result_ynan2:
                        scf.yield_(nan, nan)
                    with else_(result_ynan2):
                        with if_(x_is_zero, results=[FTy, FTy]) as result_xzero:
                            abs_y = math_dialect.absf(y)
                            abs_y_gt_1 = arith.cmpf(arith.CmpFPredicate.OGT, abs_y, c1)
                            one_plus_y = arith.addf(c1, y)
                            one_minus_y = arith.subf(c1, y)
                            half = arith.constant(0.5, FTy)
                            atanh_y = arith.mulf(
                                half,
                                math_dialect.log(arith.divf(one_plus_y, one_minus_y)),
                            )
                            y_plus_1 = arith.addf(y, c1)
                            y_minus_1 = arith.subf(y, c1)
                            atanh_inv_y = arith.mulf(
                                half,
                                math_dialect.log(arith.divf(y_plus_1, y_minus_1)),
                            )
                            real_big = math_dialect.copysign(pi_over_2, x)
                            real_small = x
                            real_res = arith.select(abs_y_gt_1, real_big, real_small)
                            imag_res = arith.select(abs_y_gt_1, atanh_inv_y, atanh_y)
                            scf.yield_(real_res, imag_res)
                        with else_(result_xzero):
                            one_plus_iz_real = arith.subf(c1, y)
                            one_plus_iz_imag = x
                            one_minus_iz_real = arith.addf(c1, y)
                            one_minus_iz_imag = arith.negf(x)
                            denom = arith.addf(
                                arith.mulf(one_minus_iz_real, one_minus_iz_real),
                                arith.mulf(one_minus_iz_imag, one_minus_iz_imag),
                            )
                            ratio_real = arith.divf(
                                arith.addf(
                                    arith.mulf(one_plus_iz_real, one_minus_iz_real),
                                    arith.mulf(one_plus_iz_imag, one_minus_iz_imag),
                                ),
                                denom,
                            )
                            ratio_imag = arith.divf(
                                arith.subf(
                                    arith.mulf(one_plus_iz_imag, one_minus_iz_real),
                                    arith.mulf(one_plus_iz_real, one_minus_iz_imag),
                                ),
                                denom,
                            )
                            log_mod = math_dialect.log(
                                math_dialect.sqrt(
                                    arith.addf(
                                        arith.mulf(ratio_real, ratio_real),
                                        arith.mulf(ratio_imag, ratio_imag),
                                    )
                                )
                            )
                            log_phase = math_dialect.atan2(ratio_imag, ratio_real)
                            half = arith.constant(0.5, FTy)
                            final_real = arith.mulf(half, log_phase)
                            raw_imag = arith.mulf(half, log_mod)
                            neg_imag = arith.negf(raw_imag)
                            # Preserve sign of y when result is zero (signed zero)
                            imag_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, neg_imag, c0)
                            signed_zero = math_dialect.copysign(c0, y)
                            final_imag = arith.select(imag_is_zero, signed_zero, neg_imag)
                            scf.yield_(final_real, final_imag)
                        scf.yield_(result_xzero.results)
                    scf.yield_(result_ynan2.results)
                scf.yield_(result_xnan.results)
            scf.yield_(result_xinf.results)
        scf.yield_(result_yinf.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_asin(z_real: FTy, z_imag: FTy):
    """asin(z) using Kahan's algorithm with real arithmetic (matches CPython).

    Uses: s = hypot(1+|x|, |y|), t = hypot(1-|x|, |y|), a = (s+t)/2
    real = copysign(asin(|x|/a), x)
    imag = copysign(log(a + sqrt(a^2-1)), y)
    """
    c0 = _zero(FTy)
    c1 = arith.constant(1.0, FTy)
    nan = _nan(FTy)
    inf = _inf(FTy)
    pi_over_2 = arith.constant(1.5707963267948966, FTy)

    x = z_real
    y = z_imag

    x_isnan = math_dialect.isnan(x)
    y_isnan = math_dialect.isnan(y)
    x_isinf = math_dialect.isinf(x)
    y_isinf = math_dialect.isinf(y)
    y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, y, c0)

    with if_(y_isinf, results=[FTy, FTy]) as result:
        with if_(x_isnan, results=[FTy, FTy]) as result_xnan:
            scf.yield_(x, y)
        with else_(result_xnan):
            real_res = math_dialect.copysign(pi_over_2, x)
            scf.yield_(real_res, y)
        scf.yield_(result_xnan.results)
    with else_(result):
        with if_(x_isnan, results=[FTy, FTy]) as result_xnan2:
            with if_(y_is_zero, results=[FTy, FTy]) as result_yz:
                scf.yield_(nan, y)
            with else_(result_yz):
                scf.yield_(nan, nan)
            scf.yield_(result_yz.results)
        with else_(result_xnan2):
            with if_(y_isnan, results=[FTy, FTy]) as result_ynan:
                with if_(x_isinf, results=[FTy, FTy]) as result_xi:
                    scf.yield_(nan, x)
                with else_(result_xi):
                    scf.yield_(nan, nan)
                scf.yield_(result_xi.results)
            with else_(result_ynan):
                with if_(x_isinf, results=[FTy, FTy]) as result_xinf:
                    real_res = math_dialect.copysign(pi_over_2, x)
                    imag_res = math_dialect.copysign(inf, y)
                    scf.yield_(real_res, imag_res)
                with else_(result_xinf):
                    # Kahan's algorithm using only real arithmetic
                    ax = math_dialect.absf(x)
                    ay = math_dialect.absf(y)

                    # s = hypot(1+ax, ay), t = hypot(1-ax, ay)
                    opax = arith.addf(c1, ax)
                    omax = arith.subf(c1, ax)
                    ay2 = arith.mulf(ay, ay)
                    s = math_dialect.sqrt(arith.addf(arith.mulf(opax, opax), ay2))
                    t = math_dialect.sqrt(arith.addf(arith.mulf(omax, omax), ay2))

                    # a = (s+t)/2, b = ax/a
                    half = arith.constant(0.5, FTy)
                    a = arith.mulf(half, arith.addf(s, t))
                    b = arith.divf(ax, a)

                    # real part = asin(b), clamped to [0, pi/2] via min(b, 1)
                    b_clamped = arith.minimumf(b, c1)
                    real_abs = math_dialect.asin(b_clamped)
                    final_real = math_dialect.copysign(real_abs, x)

                    # imag part = log(a + sqrt(a^2-1))
                    a2m1 = arith.subf(arith.mulf(a, a), c1)
                    # Ensure non-negative (numerical noise near a=1)
                    a2m1_safe = arith.maximumf(a2m1, c0)
                    imag_abs = math_dialect.log(arith.addf(a, math_dialect.sqrt(a2m1_safe)))
                    final_imag = math_dialect.copysign(imag_abs, y)

                    scf.yield_(final_real, final_imag)
                scf.yield_(result_xinf.results)
            scf.yield_(result_ynan.results)
        scf.yield_(result_xnan2.results)
    return result.results[0], result.results[1]


@func.func(sym_visibility="private", generics=[FTy])
def cmath_asinh(z_real: FTy, z_imag: FTy):
    """asinh(z) using Kahan's algorithm: asinh(x+iy) via asin(-y+ix) rotated.

    Uses same real-arithmetic Kahan formula as asin, applied to iz=(-y,x),
    then rotates result: asinh(z) = (asin_imag, -asin_real).
    """
    c0 = _zero(FTy)
    c1 = arith.constant(1.0, FTy)
    nan = _nan(FTy)
    inf = _inf(FTy)
    pi_over_2 = arith.constant(1.5707963267948966, FTy)
    pi_over_4 = arith.constant(0.7853981633974483, FTy)

    x = z_real
    y = z_imag

    x_isnan = math_dialect.isnan(x)
    y_isnan = math_dialect.isnan(y)
    x_isinf = math_dialect.isinf(x)
    y_isinf = math_dialect.isinf(y)

    with if_(x_isinf, results=[FTy, FTy]) as result:
        with if_(y_isnan, results=[FTy, FTy]) as result_ynan:
            scf.yield_(x, nan)
        with else_(result_ynan):
            with if_(y_isinf, results=[FTy, FTy]) as result_yinf:
                real_res = math_dialect.copysign(inf, x)
                imag_res = math_dialect.copysign(pi_over_4, y)
                scf.yield_(real_res, imag_res)
            with else_(result_yinf):
                real_res = math_dialect.copysign(inf, x)
                imag_res = math_dialect.copysign(c0, y)
                scf.yield_(real_res, imag_res)
            scf.yield_(result_yinf.results)
        scf.yield_(result_ynan.results)
    with else_(result):
        with if_(y_isinf, results=[FTy, FTy]) as result_yinf2:
            with if_(x_isnan, results=[FTy, FTy]) as result_xnan:
                scf.yield_(nan, y)
            with else_(result_xnan):
                real_res = math_dialect.copysign(inf, x)
                imag_res = math_dialect.copysign(pi_over_2, y)
                scf.yield_(real_res, imag_res)
            scf.yield_(result_xnan.results)
        with else_(result_yinf2):
            with if_(x_isnan, results=[FTy, FTy]) as result_xnan2:
                y_is_zero = arith.cmpf(arith.CmpFPredicate.OEQ, y, c0)
                with if_(y_is_zero, results=[FTy, FTy]) as result_yz:
                    scf.yield_(x, y)
                with else_(result_yz):
                    scf.yield_(nan, nan)
                scf.yield_(result_yz.results)
            with else_(result_xnan2):
                with if_(y_isnan, results=[FTy, FTy]) as result_ynan2:
                    scf.yield_(nan, nan)
                with else_(result_ynan2):
                    # Kahan for asin(iz) where iz=(-y, x), then rotate
                    # asin input: real=-y, imag=x
                    ay_asin = math_dialect.absf(y)  # |re(iz)| = |-y| = |y|
                    ax_asin = math_dialect.absf(x)  # |im(iz)| = |x|

                    opay = arith.addf(c1, ay_asin)
                    omay = arith.subf(c1, ay_asin)
                    ax2 = arith.mulf(ax_asin, ax_asin)
                    half = arith.constant(0.5, FTy)
                    s = math_dialect.sqrt(arith.addf(arith.mulf(opay, opay), ax2))
                    t = math_dialect.sqrt(arith.addf(arith.mulf(omay, omay), ax2))
                    a = arith.mulf(half, arith.addf(s, t))
                    b = arith.divf(ay_asin, a)

                    b_clamped = arith.minimumf(b, c1)
                    asin_real_abs = math_dialect.asin(b_clamped)
                    # asin(iz).real has sign of re(iz) = -y
                    neg_y = arith.negf(y)
                    asin_real = math_dialect.copysign(asin_real_abs, neg_y)

                    a2m1 = arith.subf(arith.mulf(a, a), c1)
                    a2m1_safe = arith.maximumf(a2m1, c0)
                    asin_imag_abs = math_dialect.log(arith.addf(a, math_dialect.sqrt(a2m1_safe)))
                    # asin(iz).imag has sign of im(iz) = x
                    asin_imag = math_dialect.copysign(asin_imag_abs, x)

                    # asinh(z) = -i * asin(iz) = (asin_imag, -asin_real)
                    final_real = asin_imag
                    final_imag = arith.negf(asin_real)
                    scf.yield_(final_real, final_imag)
                scf.yield_(result_ynan2.results)
            scf.yield_(result_xnan2.results)
        scf.yield_(result_yinf2.results)
    return result.results[0], result.results[1]


def get_cmath_intrinsics_module():
    with mlir_mod_ctx() as ctx:
        for fty in [T.f32(), T.f64()]:
            for fn in [
                cmath_exp,
                cmath_sinh,
                cmath_cosh,
                cmath_sin,
                cmath_cos,
                cmath_rect,
                cmath_sqrt,
                cmath_acos,
                cmath_acosh,
                cmath_atan,
                cmath_asin,
                cmath_asinh,
            ]:
                f = fn[fty]
                f.func_attrs["alwaysinline"] = UnitAttr.get()
                f.emit()
    return str(ctx.module)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import overload
from numba_cuda_mlir.numba_cuda import typing
import inspect
from numba_cuda_mlir import types
from functools import singledispatch

import importlib.util
import numba_cuda_mlir._mlir.ir as ir
from numba_cuda_mlir._mlir.dialects import llvm
from numba_cuda_mlir._mlir.extras import types as T
import numpy as np
import ctypes


@singledispatch
def to_numba_type(obj):
    raise TypeError(f"No conversion found for type {obj}")


@to_numba_type.register(ir.Value)
def _(val: ir.Value) -> types.Type:
    """Convert an MLIR Value to a Numba type by extracting its type attribute."""
    return to_numba_type(val.type)


@to_numba_type.register(inspect.Signature)
def numba_sig_from_pyfunc(pysig: inspect.Signature) -> typing.Signature:
    args = list(pysig.parameters.values())
    if not all(hasattr(arg, "annotation") for arg in args):
        raise ValueError("All arguments must have type annotations")
    argtypes = [arg.annotation for arg in args]

    if not all(isinstance(argtype, types.Type) for argtype in argtypes):
        raise TypeError(
            "All arguments must have type annotations of type numba_cuda_mlir.types.Type"
        )

    retty = pysig.return_annotation
    if retty is inspect._empty:
        retty = types.none
    if retty and not isinstance(retty, types.NoneType) and retty is not None:
        raise TypeError("Return type must be None or numba_cuda_mlir.types.none")

    return typing.signature(types.none, *argtypes)


@to_numba_type.register(ir.TypeAttr)
def _(ty: ir.TypeAttr) -> types.Signature:
    return to_numba_type(ty.value)


@to_numba_type.register(ir.Type)
@to_numba_type.register(ir.FunctionType)
def _(ty: ir.Type | ir.FunctionType) -> types.Type | typing.Signature:
    match ty:
        case ty if str(ty) == "!llvm.ptr":
            return types.CPointer(types.void)
        case ir.MemRefType():
            dtype = to_numba_type(ty.element_type)
            return types.Array(dtype, ty.rank, "A")
        case ir.VectorType():
            from numba_cuda_mlir.type_defs.vector_types import VectorType

            elem_type = to_numba_type(ty.element_type)
            return VectorType(elem_type, tuple(ty.shape))
        case ir.IntegerType():
            match ty.width:
                case 1:
                    return types.bool
                case 8:
                    return types.int8
                case 16:
                    return types.int16
                case 32:
                    return types.int32
                case 64:
                    return types.int64
                case _:
                    raise NotImplementedError(f"Not implemented for type {ty}")
        case ir.IndexType():
            return types.int64
        case ir.BF16Type():
            from numba_cuda_mlir.numba_cuda.types.ext_types import bfloat16

            return bfloat16
        case ir.Float4E2M1FNType():
            return types.f4E2M1FN
        case ir.Float6E2M3FNType():
            return types.f6E2M3FN
        case ir.Float6E3M2FNType():
            return types.f6E3M2FN
        case ir.Float8E3M4Type():
            return types.f8E3M4
        case ir.Float8E4M3B11FNUZType():
            return types.f8E4M3B11FNUZ
        case ir.Float8E4M3FNType():
            return types.f8E4M3FN
        case ir.Float8E4M3FNUZType():
            return types.f8E4M3FNUZ
        case ir.Float8E4M3Type():
            return types.f8E4M3
        case ir.Float8E5M2FNUZType():
            return types.f8E5M2FNUZ
        case ir.Float8E5M2Type():
            return types.f8E5M2
        case ir.Float8E8M0FNUType():
            return types.f8E8M0FNU
        case ir.FloatTF32Type():
            return types.tf32
        case ir.FloatType():
            match ty.width:
                case 16:
                    return types.float16
                case 32:
                    return types.float32
                case 64:
                    return types.float64
                case _:
                    raise NotImplementedError(f"Not implemented for type {ty}")
        case ir.FunctionType():
            result_type = types.void
            if len(ty.results) == 1:
                result_type = to_numba_type(ty.results[0])
            elif len(ty.results) > 1:
                # Multiple return values -> create a Tuple type
                result_types = [to_numba_type(r) for r in ty.results]
                # Check if all types are the same for UniTuple
                if all(t == result_types[0] for t in result_types):
                    result_type = types.UniTuple(result_types[0], len(result_types))
                else:
                    result_type = types.Tuple(result_types)
            return typing.signature(result_type, *[to_numba_type(arg) for arg in ty.inputs])
        case ir.Type():
            # we need to downcast LLVM types for some reason...
            # TODO(ajm): this has been fixed, need to update this everywhere
            tystr = str(ty)
            if "llvm.ptr" in tystr:
                return types.ptr
            elif "llvm.struct" in tystr:
                st = llvm.StructType(ty)
                assert not st.opaque, f"Struct type {st.name} is opaque, how did this happen?"
                name = st.name
                named_fe_type = types.AggregateType.get_named_type(name)
                assert named_fe_type is not None, f"No named type found for {name}"
                return named_fe_type
        case _:
            raise NotImplementedError(f"Not implemented for type {ty}")


@to_numba_type.register(type)
def _(obj: type) -> types.Type:
    if ctypes._SimpleCData in obj.mro():
        return ctypes_type_to_numba_type(obj)
    if issubclass(obj, np.generic):
        return np_dtype_to_numba_dtype(np.dtype(obj))
    raise NotImplementedError(f"Not implemented for type {obj}")


def ctypes_type_to_numba_type(obj: ctypes._SimpleCData) -> types.Type:
    match obj:
        case ctypes.c_void_p:
            return types.ptr
        case ctypes.c_bool:
            return types.bool
        case ctypes.c_int8:
            return types.int8
        case ctypes.c_uint8:
            return types.uint8
        case ctypes.c_int16:
            return types.int16
        case ctypes.c_uint16:
            return types.uint16
        case ctypes.c_int32:
            return types.int32
        case ctypes.c_uint32:
            return types.uint32
        case ctypes.c_int64:
            return types.int64
        case ctypes.c_uint64:
            return types.uint64
        case ctypes.c_float:
            return types.float32
        case ctypes.c_double:
            return types.float64
        case _:
            raise NotImplementedError(f"Not implemented for type {obj}")


# Add overloads, so IDEs that use pyright does not mark the code after the
# function call as unreachable.
@overload
def to_mlir_type(obj: types.Type) -> ir.Type: ...
@overload
def to_mlir_type(obj: ctypes._SimpleCData) -> ir.Type: ...
@overload
def to_mlir_type(obj: np.dtype) -> ir.Type: ...
@overload
def to_mlir_type(obj: typing.Signature) -> ir.FunctionType: ...
@overload
def to_mlir_type(obj: ir.Value) -> ir.Type: ...
@overload
def to_mlir_type(obj: type) -> ir.Type: ...


@singledispatch
def to_mlir_type(obj):
    raise TypeError(f"No conversion found for type {type(obj)}")


def to_mlir_storage_type(obj):
    from numba_cuda_mlir.models import mlir_data_manager

    if isinstance(obj, types.Type):
        return mlir_data_manager.lookup(obj).get_data_type()
    if isinstance(obj, np.dtype):
        return mlir_data_manager.lookup(to_numba_type(obj)).get_data_type()
    return to_mlir_type(obj)


def to_mlir_argument_type(obj):
    from numba_cuda_mlir.models import mlir_data_manager

    if isinstance(obj, types.Type):
        return mlir_data_manager.lookup(obj).get_argument_type()
    return to_mlir_type(obj)


def to_mlir_return_type(obj):
    from numba_cuda_mlir.models import mlir_data_manager

    if isinstance(obj, types.Type):
        return mlir_data_manager.lookup(obj).get_return_type()
    return to_mlir_type(obj)


@to_mlir_type.register(ir.Value)
def _(val: ir.Value) -> ir.Type:
    """Extract the MLIR type from an MLIR Value (including ScalarValue)."""
    return val.type


@to_mlir_type.register(type)
def _(obj: type) -> ir.Type:
    mro = obj.mro()
    if ctypes._SimpleCData in mro:
        return ctypes_type_to_mlir_type(obj)
    elif issubclass(obj, np.generic):
        return np_dtype_to_mlir_type(obj)
    raise NotImplementedError(f"Not implemented for type {obj}")


def ctypes_type_to_mlir_type(obj: ctypes._SimpleCData) -> ir.Type:
    """
    TODO(ajm): verify signedness matches user's expectations.
    NOTE: The arith dialect is exclusively signless, so it is on us to maintain the
    signedness of the types in the numba type system.
    """
    from numba_cuda_mlir._mlir.dialects import llvm

    match obj:
        case ctypes.c_void_p:
            return llvm.PointerType.get()
        case ctypes.c_bool:
            return T.bool()
        case ctypes.c_int8:
            return T.i8()
        case ctypes.c_uint8:
            return T.ui8()
        case ctypes.c_int16:
            return T.i16()
        case ctypes.c_uint16:
            return T.i16()
        case ctypes.c_int32:
            return T.i32()
        case ctypes.c_uint32:
            return T.i32()
        case ctypes.c_int64:
            return T.i64()
        case ctypes.c_uint64:
            return T.i64()
        case ctypes.c_float:
            return T.f32()
        case ctypes.c_double:
            return T.f64()
        case ctypes.POINTER():
            raise NotImplementedError(f"Not implemented for type {type(obj)}")
        case _:
            raise NotImplementedError(f"Not implemented for type {type(obj)}")


@to_mlir_type.register(typing.Signature)
def _(ty: typing.Signature) -> ir.FunctionType:
    if ty.return_type is types.void:
        return_types = []
    elif isinstance(ty.return_type, types.UniTuple):
        # Multiple returns of same type
        elem_type = to_mlir_type(ty.return_type.dtype)
        return_types = [elem_type] * ty.return_type.count
    elif isinstance(ty.return_type, types.Tuple):
        # Multiple returns of different types
        return_types = [to_mlir_type(t) for t in ty.return_type.types]
    else:
        return_types = [to_mlir_type(ty.return_type)]
    arg_types = tuple(to_mlir_type(arg_type) for arg_type in ty.args)
    return ir.FunctionType.get(results=return_types, inputs=arg_types)


@to_mlir_type.register(np.dtype)
def np_dtype_to_mlir_type(dtype: np.dtype) -> ir.Type:
    match dtype:
        case np.int8:
            return T.i8()
        case np.int16:
            return T.i16()
        case np.int32:
            return T.i32()
        case np.int64:
            return T.i64()
        case np.uint8:
            return T.i8()
        case np.uint16:
            return T.i16()
        case np.uint32:
            return T.i32()
        case np.uint64:
            return T.i64()
        case np.bool:
            return T.bool()
        case np.float16:
            return T.f16()
        case np.float32:
            return T.f32()
        case np.float64:
            return T.f64()
        case np.complex64:
            return T.complex(T.f32())
        case np.complex128:
            return T.complex(T.f64())
        case _ if dtype.kind in ("M", "m"):
            return T.i64()
        case _:
            raise ValueError(f"Cannot convert dtype {dtype} to MLIR type.")


@to_mlir_type.register(types.Type)
def _(ty: types.Type) -> ir.Type:
    from numba_cuda_mlir.type_defs import float_types
    from numba_cuda_mlir.numba_cuda.types.ext_types import (
        Bfloat16,
        GridGroup as GridGroupClass,
    )

    match ty:
        case GridGroupClass():
            return T.i64()
        case types.BooleanLiteral():
            return T.bool()
        case types.bool:
            return T.bool()
        case float_types.BFloat16Type():
            return T.bf16()
        case Bfloat16():
            return T.bf16()
        case _ if isinstance(ty, float_types.SpecialFloatType) or isinstance(
            ty,
            (
                type(types._type_fp8_e5m2),
                type(types._type_fp8_e4m3),
                type(types._type_fp8_e8m0),
                types.bfloat16_raw_class,
            ),
        ):
            from numba_cuda_mlir.models import mlir_data_manager

            return mlir_data_manager.lookup(ty).get_value_type()
        case types.float16:
            return T.f16()
        case types.float32:
            return T.f32()
        case types.float64:
            return T.f64()
        case types.int8 | types.uint8:
            return T.i8()
        case types.int16 | types.uint16:
            return T.i16()
        case types.int32 | types.uint32:
            return T.i32()
        case types.int64 | types.uint64:
            return T.i64()
        case types.NPDatetime() | types.NPTimedelta():
            return T.i64()
        case types.CPointer():
            return llvm.PointerType.get()
        case types.Complex() as ct:
            inner_type = to_mlir_type(ct.underlying_float)
            return T.complex(inner_type)
        case types.IntegerLiteral():
            return T.i64()
        case types.Array():
            from numba_cuda_mlir.types import Record

            dyn = ir.MemRefType.get_dynamic_size()
            dyn_stride = ir.MemRefType.get_dynamic_stride_or_offset()
            layout = ir.StridedLayoutAttr.get(
                offset=dyn_stride,
                strides=[dyn_stride] * ty.ndim,
            )
            # For Record arrays, use byte-based memref (i8) since llvm.ptr
            # is not a valid memref element type
            if isinstance(ty.dtype, Record):
                return ir.MemRefType.get([dyn] * ty.ndim, T.i8(), layout=layout)
            from numba_cuda_mlir.models import mlir_data_manager

            dtype = mlir_data_manager.lookup(ty.dtype).get_data_type()
            return ir.MemRefType.get([dyn] * ty.ndim, dtype, layout=layout)
        case types.AggregateType():
            st = llvm.StructType.get_identified(ty.name)
            if st.opaque:
                # Check if this is a bitfield struct
                if ty.is_bitfield_struct:
                    # Bitfield struct: get the storage type and convert to MLIR
                    storage_type = ty.get_bitfield_storage_type()
                    mlir_types = [to_mlir_type(storage_type)]
                else:
                    # Regular struct: create a field for each struct field
                    numba_types = [i[1] for i in ty.fields if i[0] is not None]
                    mlir_types = [to_mlir_type(ty) for ty in numba_types]
                st = llvm.StructType.new_identified(ty.name, mlir_types)
            return st
        case types.UnionType():
            # Union is represented as an integer of the size of the largest variant
            # Use the size_bits property which handles all variant types
            max_bits = ty.size_bits

            # Return integer type of appropriate size
            if max_bits <= 8:
                return T.i8()
            elif max_bits <= 16:
                return T.i16()
            elif max_bits <= 32:
                return T.i32()
            elif max_bits <= 64:
                return T.i64()
            else:
                raise NotImplementedError(
                    f"Union '{ty.name}' requires {max_bits} bits but only up to 64-bit unions are currently supported"
                )
        case types.EnumMember() | types.IntEnumMember():
            return to_mlir_type(ty.dtype)
        case _:
            from numba_cuda_mlir.type_defs.vector_types import VectorType

            if isinstance(ty, VectorType):
                elem_type = to_mlir_type(ty.dtype)
                return ir.VectorType.get(list(ty.shape), elem_type)

            try:
                from numba_cuda_mlir.models import mlir_data_manager

                model = mlir_data_manager.lookup(ty)
                return model.get_value_type()
            except KeyError:
                pass

            raise TypeError(f"Unsupported MLIR dtype: {ty}")


@to_mlir_type.register
def _typeref_to_mlir(ty) -> ir.Type:
    """Handle TypeRef by extracting and converting its instance_type."""
    from numba_cuda_mlir.numba_cuda.types.abstract import TypeRef

    if isinstance(ty, TypeRef):
        return to_mlir_type(ty.instance_type)
    raise TypeError(f"Expected TypeRef, got {type(ty)}")


# Explicitly register TypeRef to avoid singledispatch issues with subclassing
try:
    from numba_cuda_mlir.numba_cuda.types.abstract import TypeRef

    to_mlir_type.register(TypeRef, _typeref_to_mlir)
except ImportError:
    pass


@to_numba_type.register(np.dtype)
def np_dtype_to_numba_dtype(dtype: np.dtype) -> types.Type:
    match dtype:
        case np.bool:
            return types.bool
        case np.float16:
            return types.float16
        case np.float32:
            return types.float32
        case np.float64:
            return types.float64
        case np.int8:
            return types.int8
        case np.int16:
            return types.int16
        case np.int32:
            return types.int32
        case np.int64:
            return types.int64
        case np.uint8:
            return types.uint8
        case np.uint16:
            return types.uint16
        case np.uint32:
            return types.uint32
        case np.uint64:
            return types.uint64
        case np.complex64:
            return types.complex64
        case np.complex128:
            return types.complex128
        case _ if dtype.kind == "M":
            return types.NPDatetime(dtype.str[4:-1] if "[" in dtype.str else "")
        case _ if dtype.kind == "m":
            return types.NPTimedelta(dtype.str[4:-1] if "[" in dtype.str else "")
        case _:
            raise TypeError(f"Unsupported numpy dtype: {dtype}")


def integer_of_width(width: int) -> types.Type:
    match width:
        case 1:
            return types.bool
        case 8:
            return types.int8
        case 16:
            return types.int16
        case 32:
            return types.int32
        case 64:
            return types.int64
        case _:
            raise ValueError(f"Unsupported integer width: {width}")


def float_of_width(width: int) -> types.Type:
    match width:
        case 16:
            # TODO: this is ambiguous... as long as we don't allow bad conversions between
            # exotic float types, this won't be a problem.
            return types.float16
        case 32:
            return types.float32
        case 64:
            return types.float64
        case _:
            raise ValueError(f"Unsupported float width: {width}")


if importlib.util.find_spec("torch") is not None:
    import torch

    @to_mlir_type.register(torch.dtype)
    def torch_dtype_to_mlir_dtype(dtype: torch.dtype) -> ir.Type:
        return to_mlir_type(to_numba_type(dtype))

    @to_numba_type.register(torch.dtype)
    def torch_dtype_to_numba_dtype(dtype: torch.dtype) -> types.Type:
        from numba_cuda_mlir.type_defs import float_types

        match dtype:
            case torch.bool:
                return types.bool
            case torch.bfloat16:
                return float_types.BFloat16Type()
            case torch.float16:
                return types.float16
            case torch.float32:
                return types.float32
            case torch.float64:
                return types.float64
            case torch.int8:
                return types.int8
            case torch.int16:
                return types.int16
            case torch.int32:
                return types.int32
            case torch.int64:
                return types.int64


def inline_ptx_type_constraint_to_numba_type(constraint: str) -> types.Type:
    if len(constraint) != 1:
        raise ValueError(f"Invalid inline ptx type constraint: {constraint}")
    numba_types = {
        "h": types.int16,
        "r": types.int32,
        "l": types.int64,
        "f": types.float32,
        "d": types.float64,
        # "q": types.int128, # TODO
        "C": types.ptr,
    }
    if constraint not in numba_types:
        raise ValueError(f"Invalid inline ptx type constraint: {constraint}")
    return numba_types[constraint]


def inline_ptx_type_constraint_to_mlir_type(constraint: str) -> ir.Type:
    if len(constraint) != 1:
        raise ValueError(f"Invalid inline ptx type constraint: {constraint}")
    mlir_types = {
        "h": T.i16(),
        "r": T.i32(),
        "l": T.i64(),
        "f": T.f32(),
        "d": T.f64(),
        # "q": T.i128(),
        "C": llvm.PointerType.get(),
    }
    if constraint not in mlir_types:
        raise ValueError(f"Invalid inline ptx type constraint: {constraint}")
    return mlir_types[constraint]



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from numba_cuda_mlir import numba_cuda as cuda
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.core.errors import (
    RequireLiteralValue,
    TypingError,
    NumbaTypeError,
)
from numba_cuda_mlir.numba_cuda.typing import signature
from numba_cuda_mlir.numba_cuda.extending import overload_attribute, overload_method
from numba_cuda_mlir.numba_cuda.extending import intrinsic


# Each @intrinsic below provides the typing (the returned signature plus literal
# / type validation) that the MLIR type-inference pass needs. The codegen used
# to build NVVM IR with llvmlite, but on the MLIR path these grid / syncthreads
# / warp shuffle / warp vote operations are lowered by
# numba_cuda_mlir.lowering.cuda, so the codegen closures are never invoked (a
# numba_cuda_mlir.numba_cuda intrinsic builder is filtered out by MLIRLower).
# The codegen is therefore a single shared tombstone.
def _dead_codegen(context, builder, sig, args):
    raise NotImplementedError(
        "this intrinsic is lowered by numba_cuda_mlir.lowering.cuda on the MLIR "
        "path; the vendored llvmlite codegen is never invoked"
    )


# -------------------------------------------------------------------------------
# Grid functions


def _type_grid_function(ndim):
    val = ndim.literal_value
    if val == 1:
        restype = types.int64
    elif val in (2, 3):
        restype = types.UniTuple(types.int64, val)
    else:
        raise ValueError("argument can only be 1, 2, 3")

    return signature(restype, types.int32)


@intrinsic
def grid(typingctx, ndim):
    """grid(ndim)

    Return the absolute position of the current thread in the entire grid of
    blocks.  *ndim* should correspond to the number of dimensions declared when
    instantiating the kernel. If *ndim* is 1, a single integer is returned.
    If *ndim* is 2 or 3, a tuple of the given number of integers is returned.

    Computation of the first integer is as follows::

        cuda.threadIdx.x + cuda.blockIdx.x * cuda.blockDim.x

    and is similar for the other two indices, but using the ``y`` and ``z``
    attributes.
    """

    if not isinstance(ndim, types.IntegerLiteral):
        raise RequireLiteralValue(ndim)

    sig = _type_grid_function(ndim)

    return sig, _dead_codegen


@intrinsic
def gridsize(typingctx, ndim):
    """gridsize(ndim)

    Return the absolute size (or shape) in threads of the entire grid of
    blocks. *ndim* should correspond to the number of dimensions declared when
    instantiating the kernel. If *ndim* is 1, a single integer is returned.
    If *ndim* is 2 or 3, a tuple of the given number of integers is returned.

    Computation of the first integer is as follows::

        cuda.blockDim.x * cuda.gridDim.x

    and is similar for the other two indices, but using the ``y`` and ``z``
    attributes.
    """

    if not isinstance(ndim, types.IntegerLiteral):
        raise RequireLiteralValue(ndim)

    sig = _type_grid_function(ndim)

    return sig, _dead_codegen


@intrinsic
def _warpsize(typingctx):
    sig = signature(types.int32)

    return sig, _dead_codegen


@overload_attribute(types.Module(cuda), "warpsize", target="cuda")
def cuda_warpsize(mod):
    """
    The size of a warp. All architectures implemented to date have a warp size
    of 32.
    """

    def get(mod):
        return _warpsize()

    return get


# -------------------------------------------------------------------------------
# syncthreads


@intrinsic
def syncthreads(typingctx):
    """
    Synchronize all threads in the same thread block.  This function implements
    the same pattern as barriers in traditional multi-threaded programming: this
    function waits until all threads in the block call it, at which point it
    returns control to all its callers.
    """
    sig = signature(types.none)

    return sig, _dead_codegen


def _syncthreads_predicate(typingctx, predicate, fname):
    if not isinstance(predicate, types.Integer):
        return None

    sig = signature(types.i4, types.i4)

    return sig, _dead_codegen


@intrinsic
def syncthreads_count(typingctx, predicate):
    """
    syncthreads_count(predicate)

    An extension to numba.cuda.syncthreads where the return value is a count
    of the threads where predicate is true.
    """
    fname = "llvm.nvvm.barrier0.popc"
    return _syncthreads_predicate(typingctx, predicate, fname)


@intrinsic
def syncthreads_and(typingctx, predicate):
    """
    syncthreads_and(predicate)

    An extension to numba.cuda.syncthreads where 1 is returned if predicate is
    true for all threads or 0 otherwise.
    """
    fname = "llvm.nvvm.barrier0.and"
    return _syncthreads_predicate(typingctx, predicate, fname)


@intrinsic
def syncthreads_or(typingctx, predicate):
    """
    syncthreads_or(predicate)

    An extension to numba.cuda.syncthreads where 1 is returned if predicate is
    true for any thread or 0 otherwise.
    """
    fname = "llvm.nvvm.barrier0.or"
    return _syncthreads_predicate(typingctx, predicate, fname)


@overload_method(types.Integer, "bit_count", target="cuda")
def integer_bit_count(i):
    return lambda i: cuda.popc(i)


# -------------------------------------------------------------------------------
# Warp shuffle functions
#
# References:
#
# - https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#warp-shuffle-functions
# - https://docs.nvidia.com/cuda/nvvm-ir-spec/index.html#data-movement


@intrinsic
def shfl_sync(typingctx, mask, value, src_lane):
    """
    Shuffles ``value`` across the masked warp and returns the value from
    ``src_lane``. If this is outside the warp, then the given value is
    returned.
    """
    membermask_type = mask
    mode_value = 0
    a_type = value
    b_type = src_lane
    c_value = 0x1F
    return shfl_sync_intrinsic(typingctx, membermask_type, mode_value, a_type, b_type, c_value)


@intrinsic
def shfl_up_sync(typingctx, mask, value, delta):
    """
    Shuffles ``value`` across the masked warp and returns the value from
    ``(laneid - delta)``. If this is outside the warp, then the given value is
    returned.
    """
    membermask_type = mask
    mode_value = 1
    a_type = value
    b_type = delta
    c_value = 0
    return shfl_sync_intrinsic(typingctx, membermask_type, mode_value, a_type, b_type, c_value)


@intrinsic
def shfl_down_sync(typingctx, mask, value, delta):
    """
    Shuffles ``value`` across the masked warp and returns the value from
    ``(laneid + delta)``. If this is outside the warp, then the given value is
    returned.
    """
    membermask_type = mask
    mode_value = 2
    a_type = value
    b_type = delta
    c_value = 0x1F
    return shfl_sync_intrinsic(typingctx, membermask_type, mode_value, a_type, b_type, c_value)


@intrinsic
def shfl_xor_sync(typingctx, mask, value, lane_mask):
    """
    Shuffles ``value`` across the masked warp and returns the value from
    ``(laneid ^ lane_mask)``.
    """
    membermask_type = mask
    mode_value = 3
    a_type = value
    b_type = lane_mask
    c_value = 0x1F
    return shfl_sync_intrinsic(typingctx, membermask_type, mode_value, a_type, b_type, c_value)


def shfl_sync_intrinsic(
    typingctx,
    membermask_type,
    mode_value,
    a_type,
    b_type,
    c_value,
):
    if a_type not in (types.i4, types.i8, types.f4, types.f8):
        raise TypingError("shfl_sync only supports 32- and 64-bit ints and floats")

    sig = signature(a_type, membermask_type, a_type, b_type)

    return sig, _dead_codegen


# -------------------------------------------------------------------------------
# Warp vote functions
#
# References:
#
# - https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#warp-vote-functions
# - https://docs.nvidia.com/cuda/nvvm-ir-spec/index.html?highlight=data%2520movement#vote


@intrinsic
def all_sync(typingctx, mask_type, predicate_type):
    """
    If for all threads in the masked warp the predicate is true, then
    a non-zero value is returned, otherwise 0 is returned.
    """
    mode_value = 0
    # Validate the mask / predicate types (raises on unsupported types).
    vote_sync_intrinsic(typingctx, mask_type, mode_value, predicate_type)

    sig_outer = signature(types.b1, mask_type, predicate_type)
    return sig_outer, _dead_codegen


@intrinsic
def any_sync(typingctx, mask_type, predicate_type):
    """
    If for any thread in the masked warp the predicate is true, then
    a non-zero value is returned, otherwise 0 is returned.
    """
    mode_value = 1
    vote_sync_intrinsic(typingctx, mask_type, mode_value, predicate_type)

    sig_outer = signature(types.b1, mask_type, predicate_type)
    return sig_outer, _dead_codegen


@intrinsic
def eq_sync(typingctx, mask_type, predicate_type):
    """
    If for all threads in the masked warp the boolean predicate is the same,
    then a non-zero value is returned, otherwise 0 is returned.
    """
    mode_value = 2
    vote_sync_intrinsic(typingctx, mask_type, mode_value, predicate_type)

    sig_outer = signature(types.b1, mask_type, predicate_type)
    return sig_outer, _dead_codegen


@intrinsic
def ballot_sync(typingctx, mask_type, predicate_type):
    """
    Returns a mask of all threads in the warp whose predicate is true,
    and are within the given mask.
    """
    mode_value = 3
    vote_sync_intrinsic(typingctx, mask_type, mode_value, predicate_type)

    sig_outer = signature(types.i4, mask_type, predicate_type)
    return sig_outer, _dead_codegen


def vote_sync_intrinsic(typingctx, mask_type, mode_value, predicate_type):
    # Validate mode value
    if mode_value not in (0, 1, 2, 3):
        raise ValueError("Mode must be 0 (all), 1 (any), 2 (eq), or 3 (ballot)")

    if types.unliteral(mask_type) not in types.integer_domain:
        raise NumbaTypeError(f"Mask type must be an integer. Got {mask_type}")
    predicate_types = types.integer_domain | {types.boolean}

    if types.unliteral(predicate_type) not in predicate_types:
        raise NumbaTypeError(f"Predicate must be an integer or boolean. Got {predicate_type}")

    sig = signature(types.Tuple((types.i4, types.b1)), mask_type, predicate_type)

    return sig, _dead_codegen



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
Struct support for numba_cuda_mlir - similar to ctypes structures
"""

from collections import namedtuple, Counter
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.type_defs.aggregate_types import AggregateType
from typing import Any
import ctypes
from numba_cuda_mlir.descriptor import mlir_target
from numba_cuda_mlir.numba_cuda.typing.templates import ConcreteTemplate
from numba_cuda_mlir import typing
from numba_cuda_mlir.lowering.struct import _lower_struct_construction_impl

# Counter for generating unique anonymous struct names if no name is provided
_anon_struct_counter = 0


class StructInstance:
    """
    This class holds the actual field values and can be passed to
    JIT-compiled kernels. It's similar to a ctypes Structure instance.
    """

    def __init__(self, struct_type: AggregateType, **kwargs):
        self._struct_type = struct_type
        self._fields = {}

        # Initialize all fields from kwargs (skip padding fields where field_name is None)
        for field_name, field_type, *_ in struct_type.fields:
            if field_name is None:
                # Skip padding fields - they're not accessible
                continue
            if field_name in kwargs:
                value = kwargs[field_name]
                # TODO: Add type checking/conversion here
                self._fields[field_name] = value
            else:
                # Initialize with default value (0 or equivalent)
                self._fields[field_name] = self._get_default_value(field_type)

    def _get_default_value(self, field_type: types.Type) -> Any:
        """Get a default value for a field type."""
        if isinstance(field_type, types.Integer):
            return 0
        elif isinstance(field_type, types.Float):
            return 0.0
        elif isinstance(field_type, types.Boolean):
            return False
        else:
            return None

    def __getattr__(self, name: str) -> Any:
        """Allow accessing fields via dot notation."""
        if name.startswith("_"):
            # Private attributes go through normal lookup
            return object.__getattribute__(self, name)

        if name in self._fields:
            return self._fields[name]

        raise AttributeError(f"Struct '{self._struct_type.name}' has no field '{name}'")

    def __setattr__(self, name: str, value: Any):
        """Allow setting fields via dot notation."""
        if name.startswith("_"):
            # Private attributes go through normal setattr
            object.__setattr__(self, name, value)
        elif hasattr(self, "_fields") and name in self._fields:
            # TODO: Add type checking here
            self._fields[name] = value
        else:
            object.__setattr__(self, name, value)

    def __repr__(self) -> str:
        field_strs = [
            f"{name}={self._fields[name]}"
            for name, _, _ in self._struct_type.fields
            if name is not None
        ]
        return f"{self._struct_type.name}({', '.join(field_strs)})"


class StructTypeWrapper:
    """
    Wrapper around AggregateType that adds a constructor.

    This allows users to call `MyStruct(field1=val1, ...)` to create instances.

    The underlying AggregateType can be accessed via the `_type` property,
    which is needed when passing the type to compile() or similar APIs.
    """

    def __init__(self, struct_type: AggregateType):
        self._type = struct_type  # Public property: access via MyStruct._type
        # Copy type attributes so this wrapper can be used as a type annotation
        self.name = struct_type.name
        self.fields = struct_type.fields
        # Will be set by _register_struct_constructor
        self._constructor_template = None

    def __call__(self, **kwargs) -> StructInstance:
        """Create an instance of this struct."""
        return StructInstance(self._type, **kwargs)

    def __repr__(self) -> str:
        return f"StructType({self.name})"

    # Make it work as a type for annotations and isinstance checks
    def __getattr__(self, name):
        return getattr(self._type, name)

    def __instancecheck__(self, instance):
        return isinstance(instance, self._type.__class__)

    def __subclasscheck__(self, subclass):
        return issubclass(subclass, self._type.__class__)


def struct(
    fields: list[tuple[str, types.Type, int | None]], name: str | None = None
) -> StructTypeWrapper:
    """
    Create a struct type for use in CUDA kernels.

    This creates a struct type that can be:
    - Modified in device functions (mutable)
    - Accessed via getattr/setattr
    - Support bitfields with specified bit widths
    - Support explicit padding in bitfields using None as field name

    Args:
        fields: List of tuples specifying field names, types, and optional bit widths.
                Use None as field name for padding bits in bitfields.
        name: Optional name for the struct type. If not provided, generates a unique anonymous name.

    Returns:
        A StructTypeWrapper that can be called to create instances

    Example:
        >>> from numba_cuda_mlir import struct,
        >>> from numba import types
        >>>
        >>> # Regular fields
        >>> MyStruct = struct(
        ...     "MyStruct",
        ...     [
        ...         ("a", types.int32),
        ...         ("b", types.float32),
        ...     ],
        ... )
        >>>
        >>> # Bitfields
        >>> MyBitfields = struct(
        ...     "MyBitfields",
        ...     [
        ...         ("flag", types.uint32, 1),  # 1-bit field
        ...         ("value", types.uint32, 5),  # 5-bit field
        ...     ],
        ... )
        >>>
        >>> # Bitfields with padding
        >>> MyPaddedBitfields = struct(
        ...     "MyPaddedBitfields",
        ...     [
        ...         ("a", types.uint32, 4),  # 4 bits
        ...         (None, types.uint32, 4),  # 4 bits padding
        ...         ("b", types.uint32, 8),  # 8 bits
        ...     ],
        ... )
    """
    # Generate unique name for anonymous structs to avoid registry collisions
    if name is None:
        global _anon_struct_counter
        name = f"AnonymousStruct_{_anon_struct_counter}"
        _anon_struct_counter += 1

    # Validate that field names are unique (excluding None for padding)
    field_names = [field[0] for field in fields if field[0] is not None]
    name_counts = Counter(field_names)
    duplicates = [name for name, count in name_counts.items() if count > 1]
    if duplicates:
        raise ValueError(
            f"Struct '{name}' has duplicate field names: {duplicates}. "
            f"All field names must be unique within a struct definition."
        )

    # Convert  namedtuples to (name, type, bit_width) tuples for AggregateType
    field_tuples = [(field[0], field[1], field[2] if len(field) > 2 else None) for field in fields]

    # Create the AggregateType
    struct_type = AggregateType(name, field_tuples)

    # Wrap it so it can be called to create instances
    wrapper = StructTypeWrapper(struct_type)

    # Register the constructor as callable in device code
    _register_struct_constructor(wrapper, struct_type)

    return wrapper


def _register_struct_constructor(wrapper: StructTypeWrapper, struct_type: AggregateType):
    """
    Register a struct constructor so it can be called in device code.

    This allows `MyStruct()` to work inside kernels, where it will allocate
    the struct on the stack and return it.
    """

    typingctx = mlir_target.typing_context
    targetctx = mlir_target.target_context

    # Create a signature for the constructor: () -> struct_type
    # The constructor takes no arguments and returns the struct type
    sig = typing.signature(struct_type)

    # Create a template that makes this wrapper callable
    class struct_constructor_template(ConcreteTemplate):
        key = wrapper
        cases = [sig]

    # Store the template in the wrapper so typeof_impl can access it
    wrapper._constructor_template = struct_constructor_template

    # Register it in the typing context
    typingctx.insert_user_function(wrapper, struct_constructor_template)

    def lower_struct_wrapper_call(builder, target, args, kwargs):
        """Lower calls to this specific struct constructor."""
        return _lower_struct_construction_impl(None, builder, target, args, kwargs)

    targetctx.insert_func_defn([(lower_struct_wrapper_call, wrapper, ())])



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Consteval AST transformation pass
import ast
import copy
import inspect
from typing import Callable

from numba_cuda_mlir.ast_transforms.common import get_function_context
from numba_cuda_mlir.ast_transforms.pipeline import ASTTransformPass, TransformContext


class ConstevalError(Exception):
    """Raised when a consteval expression cannot be evaluated at compile time."""

    pass


class TargetOptionsReplacer(ast.NodeTransformer):
    """Replace numba_cuda_mlir.current_target_options() calls with a context reference."""

    def visit_Call(self, node: ast.Call) -> ast.AST:
        node = self.generic_visit(node)
        # Match: numba_cuda_mlir.current_target_options() or current_target_options()
        if self._is_current_target_options_call(node):
            return ast.copy_location(
                ast.Name(id="__numba_cuda_mlir_target_options__", ctx=ast.Load()), node
            )
        return node

    def _is_current_target_options_call(self, node: ast.Call) -> bool:
        """Check if this is a call to current_target_options()."""
        if node.args or node.keywords:
            return False
        func = node.func
        # current_target_options()
        if isinstance(func, ast.Name) and func.id == "current_target_options":
            return True
        # numba_cuda_mlir.current_target_options() or cuda.current_target_options()
        if isinstance(func, ast.Attribute) and func.attr == "current_target_options":
            return True
        return False


class VariableReplacer(ast.NodeTransformer):
    """Replace a variable name with an expression throughout an AST."""

    def __init__(self, var_name: str, replacement: ast.expr):
        self.var_name = var_name
        self.replacement = replacement

    def visit_Name(self, node: ast.Name) -> ast.AST:
        if node.id == self.var_name:
            return ast.copy_location(copy.deepcopy(self.replacement), node)
        return node


class ConstevalTransformer(ast.NodeTransformer):
    """AST transformer that evaluates consteval/literally calls at compile time.

    This transformer:
    1. Tracks local constants from `name = consteval(...)` assignments
    2. Uses those constants when evaluating subsequent consteval expressions
    3. Supports chained constevals like: `a = consteval(X); if consteval(a + 5): ...`
    4. Handles complex objects by storing them in globals and referencing by name
    5. Provides access to argument types (Numba types) via parameter names
    6. Provides access to target options via numba_cuda_mlir.current_target_options()
    7. Supports `with consteval():` blocks for multi-statement compile-time execution
    """

    CONSTEVAL_NAMES = {"consteval", "literally"}
    # Types that can be safely stored in ast.Constant
    CONSTANT_TYPES = (int, float, str, bool, type(None), bytes, tuple)

    def __init__(
        self,
        func: Callable,
        targetoptions: dict = None,
        argtypes: tuple = None,
    ):
        self.func = func
        self.base_context = get_function_context(func)
        self.local_consts = {}  # Track values from consteval assignments
        self.stored_values = {}  # Complex values stored in globals
        self.modified = False
        self._value_counter = 0
        self.targetoptions = targetoptions or {}
        self.argtypes = argtypes or ()
        self.param_type_map = self._build_param_type_map(func, self.argtypes)

    def _build_param_type_map(self, func: Callable, argtypes: tuple) -> dict:
        """Build a mapping from parameter names to their Numba types."""
        sig = inspect.signature(func)
        params = list(sig.parameters.keys())
        result = {}
        for i, name in enumerate(params):
            if i < len(argtypes):
                result[name] = argtypes[i]
        return result

    @property
    def context(self):
        """Combined context: base (globals + closure) plus local constants and param types."""
        ctx = self.base_context.copy()
        ctx.update(self.local_consts)
        ctx.update(self.stored_values)
        # Add parameter types - these shadow any globals with the same name
        ctx.update(self.param_type_map)
        # Add a resolver for current_target_options()
        ctx["__numba_cuda_mlir_target_options__"] = self.targetoptions
        return ctx

    def _is_consteval_call(self, node: ast.expr) -> bool:
        """Check if a node is a call to consteval or literally."""
        if not isinstance(node, ast.Call):
            return False
        if isinstance(node.func, ast.Name):
            return node.func.id in self.CONSTEVAL_NAMES
        elif isinstance(node.func, ast.Attribute):
            return node.func.attr in self.CONSTEVAL_NAMES
        return False

    def _eval_expr(self, node: ast.expr) -> any:
        """Evaluate an AST expression node using the current context."""
        # Pre-process: replace current_target_options() calls with context reference
        node = TargetOptionsReplacer().visit(copy.deepcopy(node))
        ast.fix_missing_locations(node)
        expr_source = ast.unparse(node)
        try:
            return eval(expr_source, self.context)
        except Exception as e:
            raise ConstevalError(
                f"Cannot evaluate consteval argument '{expr_source}' at compile time: {e}"
            ) from e

    def _can_be_constant(self, value) -> bool:
        """Check if a value can be stored in an ast.Constant node."""
        if isinstance(value, self.CONSTANT_TYPES):
            # For tuples, recursively check all elements
            if isinstance(value, tuple):
                return all(self._can_be_constant(v) for v in value)
            return True
        return False

    def _store_value(self, value) -> str:
        """Store a complex value and return its reference name."""
        name = f"__consteval_{self._value_counter}__"
        self._value_counter += 1
        self.stored_values[name] = value
        return name

    def _transform_consteval(self, node: ast.Call) -> ast.AST:
        """Transform a consteval call to a Constant or Name node."""
        if len(node.args) != 1 or node.keywords:
            raise ConstevalError("consteval expects exactly one positional argument")

        arg = node.args[0]
        value = self._eval_expr(arg)
        self.modified = True

        if self._can_be_constant(value):
            return ast.copy_location(ast.Constant(value=value), node)
        else:
            # Store complex value and reference it by name
            name = self._store_value(value)
            return ast.copy_location(ast.Name(id=name, ctx=ast.Load()), node)

    def _process_statement_list(self, stmts: list[ast.stmt]) -> list[ast.stmt]:
        """Process a list of statements in order, tracking consteval assignments."""
        result = []
        for stmt in stmts:
            transformed = self._transform_statement(stmt)
            if isinstance(transformed, list):
                result.extend(transformed)
            else:
                result.append(transformed)
        return result

    def _get_consteval_value(self, node: ast.AST):
        """Get the actual value from a transformed consteval node."""
        if isinstance(node, ast.Constant):
            return node.value
        elif isinstance(node, ast.Name) and node.id in self.stored_values:
            return self.stored_values[node.id]
        return None

    def _transform_statement(self, stmt: ast.stmt) -> ast.stmt | list[ast.stmt]:
        """Transform a single statement, handling consteval assignments specially."""
        # Check for simple assignment: name = consteval(...)
        if isinstance(stmt, ast.Assign) and len(stmt.targets) == 1:
            target = stmt.targets[0]
            if isinstance(target, ast.Name) and self._is_consteval_call(stmt.value):
                # Evaluate and track this consteval assignment
                transformed = self._transform_consteval(stmt.value)
                value = self._get_consteval_value(transformed)
                if value is not None:
                    self.local_consts[target.id] = value
                stmt.value = transformed
                return stmt

        # Check for annotated assignment: name: type = consteval(...)
        if isinstance(stmt, ast.AnnAssign) and stmt.value is not None:
            if isinstance(stmt.target, ast.Name) and self._is_consteval_call(stmt.value):
                transformed = self._transform_consteval(stmt.value)
                value = self._get_consteval_value(transformed)
                if value is not None:
                    self.local_consts[stmt.target.id] = value
                stmt.value = transformed
                return stmt

        # Check for loop unrolling: for <var> in consteval(...):
        if isinstance(stmt, ast.For) and self._is_consteval_call(stmt.iter):
            return self._unroll_for_loop(stmt)

        # Check for consteval block: with consteval():
        if isinstance(stmt, ast.With) and self._is_consteval_with(stmt):
            return self._transform_consteval_block(stmt)

        # For other statements, use generic visiting
        return self.visit(stmt)

    def _unroll_for_loop(self, node: ast.For) -> list[ast.stmt]:
        """Unroll a for loop with consteval iterator."""
        self._check_unroll_loop_control(node)

        # Evaluate the iterator
        iter_value = self._eval_expr(node.iter.args[0])
        try:
            items = list(iter_value)
        except TypeError as e:
            raise ConstevalError(
                f"consteval iterator must be iterable, got {type(iter_value).__name__}"
            ) from e

        self.modified = True

        # Unroll: for each value, copy the body and replace the variable
        unrolled = []
        for value in items:
            # Save current local_consts state
            saved_consts = self.local_consts.copy()
            bindings = self._bind_loop_target(node.target, value)
            # Add loop variables to local_consts for this iteration
            self.local_consts.update(bindings)
            replacements = {}
            for var_name, bound_value in bindings.items():
                if self._can_be_constant(bound_value):
                    replacements[var_name] = ast.Constant(value=bound_value)
                else:
                    replacements[var_name] = ast.Name(
                        id=self._store_value(bound_value), ctx=ast.Load()
                    )

            for body_stmt in node.body:
                # Deep copy the statement
                stmt_copy = copy.deepcopy(body_stmt)
                # Replace each loop variable with its constant value
                for var_name, replacement in replacements.items():
                    replacer = VariableReplacer(var_name, replacement)
                    stmt_copy = replacer.visit(stmt_copy)
                ast.fix_missing_locations(stmt_copy)
                # Process the statement (handles nested constevals)
                transformed = self._transform_statement(stmt_copy)
                if isinstance(transformed, list):
                    unrolled.extend(transformed)
                else:
                    unrolled.append(transformed)

            # Restore local_consts state
            self.local_consts = saved_consts

        # Note: we ignore the else clause (orelse) since unrolled loops
        # don't have a natural "else" semantic
        return unrolled

    def _check_unroll_loop_control(self, node: ast.For) -> None:
        """Reject loop control statements that would escape an unrolled loop."""

        class LoopControlFinder(ast.NodeVisitor):
            control = None

            def visit_Break(self, node: ast.Break) -> None:
                self.control = "break"

            def visit_Continue(self, node: ast.Continue) -> None:
                self.control = "continue"

            def visit_For(self, node: ast.For) -> None:
                for stmt in node.orelse:
                    self.visit(stmt)

            def visit_AsyncFor(self, node: ast.AsyncFor) -> None:
                for stmt in node.orelse:
                    self.visit(stmt)

            def visit_While(self, node: ast.While) -> None:
                for stmt in node.orelse:
                    self.visit(stmt)

        finder = LoopControlFinder()
        for stmt in node.body:
            finder.visit(stmt)
        if finder.control:
            raise ConstevalError(f"Loop unrolling does not support {finder.control} statements")

    def _bind_loop_target(self, target: ast.expr, value) -> dict[str, object]:
        """Bind a consteval loop target to a compile-time value."""
        if isinstance(target, ast.Name):
            return {target.id: value}
        if isinstance(target, ast.Starred):
            raise ConstevalError("Loop unrolling does not support starred targets")
        if not isinstance(target, (ast.Tuple, ast.List)):
            raise ConstevalError(
                "Loop unrolling only supports name, tuple, or list targets, "
                f"got {type(target).__name__}"
            )

        try:
            values = list(value)
        except TypeError as e:
            raise ConstevalError(
                f"Cannot unpack {type(value).__name__} into {ast.unparse(target)}"
            ) from e

        if len(values) != len(target.elts):
            direction = "not enough" if len(values) < len(target.elts) else "too many"
            raise ConstevalError(
                f"{direction} values to unpack (expected {len(target.elts)}, got {len(values)})"
            )

        bindings = {}
        for element, element_value in zip(target.elts, values):
            bindings.update(self._bind_loop_target(element, element_value))
        return bindings

    def _is_consteval_with(self, node: ast.With) -> bool:
        """Check if this is a 'with consteval():' block."""
        if len(node.items) != 1:
            return False
        item = node.items[0]
        return self._is_consteval_call(item.context_expr)

    def _transform_consteval_block(self, node: ast.With) -> list[ast.stmt]:
        """Transform a 'with consteval():' block by executing it at compile time."""
        item = node.items[0]

        # Error if 'as var:' syntax is used
        if item.optional_vars is not None:
            var_name = ast.unparse(item.optional_vars)
            raise ConstevalError(
                f"'with consteval() as {var_name}:' is not supported.\n"
                "Use 'with consteval():' and extract values with consteval(var) after the block."
            )

        self.modified = True

        # Build execution context
        exec_globals = self.context.copy()
        exec_locals = {}

        # Execute each statement in the block
        for stmt in node.body:
            # Handle nested consteval blocks recursively
            if isinstance(stmt, ast.With) and self._is_consteval_with(stmt):
                # Recursively process nested block - this updates local_consts
                saved_consts = self.local_consts.copy()
                self.local_consts = dict(exec_locals)  # Use current exec state
                self.local_consts.update(exec_globals)
                self._transform_consteval_block(stmt)
                # Capture any new variables from the nested block
                for k, v in self.local_consts.items():
                    if k not in exec_globals:
                        exec_locals[k] = v
                        exec_globals[k] = v
                self.local_consts = saved_consts
                continue

            # Pre-process: replace current_target_options() calls
            stmt = TargetOptionsReplacer().visit(copy.deepcopy(stmt))
            ast.fix_missing_locations(stmt)
            stmt_source = ast.unparse(stmt)
            try:
                exec(stmt_source, exec_globals, exec_locals)
                # Update exec_globals with any new locals so subsequent statements can use them
                exec_globals.update(exec_locals)
            except Exception as e:
                raise ConstevalError(
                    f"Error executing consteval block statement:\n  {stmt_source}\nError: {e}"
                ) from e

        # Track all variables defined in the block for use in subsequent consteval() calls
        self.local_consts.update(exec_locals)

        # Return empty list - the block is removed from output
        return []

    def visit_With(self, node: ast.With) -> ast.AST | list[ast.stmt]:
        """Handle with statements, transforming consteval blocks."""
        if self._is_consteval_with(node):
            return self._transform_consteval_block(node)
        # Normal with statement - process body
        node.body = self._process_statement_list(node.body)
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        """Process function body statements in order."""
        node.body = self._process_statement_list(node.body)
        # Don't process decorators - they're not part of the function body
        return node

    def visit_If(self, node: ast.If) -> ast.If:
        """Process if statement bodies in order."""
        node.test = self.visit(node.test)
        node.body = self._process_statement_list(node.body)
        node.orelse = self._process_statement_list(node.orelse)
        return node

    def visit_For(self, node: ast.For) -> ast.For:
        """Process for loop body in order."""
        node.iter = self.visit(node.iter)
        node.body = self._process_statement_list(node.body)
        node.orelse = self._process_statement_list(node.orelse)
        return node

    def visit_While(self, node: ast.While) -> ast.While:
        """Process while loop body in order."""
        node.test = self.visit(node.test)
        node.body = self._process_statement_list(node.body)
        node.orelse = self._process_statement_list(node.orelse)
        return node

    def visit_Call(self, node: ast.Call) -> ast.AST:
        """Transform consteval calls to constants."""
        # First visit children (for nested calls)
        node = self.generic_visit(node)

        if not self._is_consteval_call(node):
            return node

        return self._transform_consteval(node)


def transform_consteval(
    func: Callable,
    tree: ast.Module,
    targetoptions: dict = None,
    argtypes: tuple = None,
) -> tuple[ast.Module, bool, dict]:
    """Transform consteval calls in the AST to literal constants.

    Args:
        func: The function being transformed
        tree: The AST to transform
        targetoptions: Target options dict from @jit decorator
        argtypes: Tuple of Numba types for the function arguments

    Returns (transformed_tree, was_modified, stored_values).
    stored_values is a dict of name -> value for complex objects that couldn't
    be stored as AST constants. These should be injected into the function's globals.
    """
    transformer = ConstevalTransformer(func, targetoptions, argtypes)
    new_tree = transformer.visit(tree)
    ast.fix_missing_locations(new_tree)
    return new_tree, transformer.modified, transformer.stored_values


class ConstevalPass(ASTTransformPass):
    """Pipeline pass that evaluates consteval/literally calls at compile time."""

    @property
    def name(self) -> str:
        return "Consteval"

    def transform(self, tree: ast.Module, context: TransformContext) -> tuple[ast.Module, bool]:
        tree, modified, stored_values = transform_consteval(
            context.func, tree, context.targetoptions, context.argtypes
        )
        context.stored_values.update(stored_values)
        return tree, modified
